@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 /*
  * A very simple font cache and rasterizer that uses FreeType
  * to draw fonts from a single OpenGL texture. The code uses
@@ -61,8 +83,8 @@ static fz_font *g_font = NULL;
 static void clear_font_cache(void)
 {
 #if PADDING > 0
-	unsigned char *zero = malloc(g_cache_w * g_cache_h);
-	memset(zero, 0, g_cache_w * g_cache_h);
+	unsigned char *zero = malloc((size_t)g_cache_w * g_cache_h);
+	memset(zero, 0, (size_t)g_cache_w * g_cache_h);
 	glBindTexture(GL_TEXTURE_2D, g_cache_tex);
 	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, g_cache_w, g_cache_h, GL_ALPHA, GL_UNSIGNED_BYTE, zero);
 	free(zero);
@@ -249,21 +271,21 @@ static float ui_draw_glyph(fz_font *font, float size, int gid, float x, float y)
 	return fz_advance_glyph(ctx, font, gid, 0) * size;
 }
 
-float ui_measure_character(int ucs)
+float ui_measure_character(int c)
 {
 	fz_font *font;
-	int gid = fz_encode_character_with_fallback(ctx, g_font, ucs, 0, 0, &font);
+	int gid = fz_encode_character_with_fallback(ctx, g_font, c, 0, 0, &font);
 	return fz_advance_glyph(ctx, font, gid, 0) * ui.fontsize;
 }
 
-float ui_draw_character(int ucs, float x, float y)
+static float ui_draw_character_imp(float x, float y, int c)
 {
 	fz_font *font;
-	int gid = fz_encode_character_with_fallback(ctx, g_font, ucs, 0, 0, &font);
+	int gid = fz_encode_character_with_fallback(ctx, g_font, c, 0, 0, &font);
 	return ui_draw_glyph(font, ui.fontsize, gid, x, y);
 }
 
-void ui_begin_text(void)
+static void ui_begin_text(void)
 {
 	glBindTexture(GL_TEXTURE_2D, g_cache_tex);
 	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
@@ -272,24 +294,23 @@ void ui_begin_text(void)
 	glBegin(GL_QUADS);
 }
 
-void ui_end_text(void)
+static void ui_end_text(void)
 {
 	glEnd();
 	glDisable(GL_TEXTURE_2D);
 	glDisable(GL_BLEND);
 }
 
-float ui_draw_string(float x, float y, const char *str)
+void ui_draw_string(float x, float y, const char *str)
 {
-	int ucs;
+	int c;
 	ui_begin_text();
 	while (*str)
 	{
-		str += fz_chartorune(&ucs, str);
-		x += ui_draw_character(ucs, x, y + ui.baseline);
+		str += fz_chartorune(&c, str);
+		x += ui_draw_character_imp(x, y + ui.baseline, c);
 	}
 	ui_end_text();
-	return x;
 }
 
 void ui_draw_string_part(float x, float y, const char *s, const char *e)
@@ -299,19 +320,26 @@ void ui_draw_string_part(float x, float y, const char *s, const char *e)
 	while (s < e)
 	{
 		s += fz_chartorune(&c, s);
-		x += ui_draw_character(c, x, y + ui.baseline);
+		x += ui_draw_character_imp(x, y + ui.baseline, c);
 	}
 	ui_end_text();
 }
 
+void ui_draw_character(float x, float y, int c)
+{
+	ui_begin_text();
+	ui_draw_character_imp(x, y + ui.baseline, c);
+	ui_end_text();
+}
+
 float ui_measure_string(const char *str)
 {
-	int ucs;
+	int c;
 	float x = 0;
 	while (*str)
 	{
-		str += fz_chartorune(&ucs, str);
-		x += ui_measure_character(ucs);
+		str += fz_chartorune(&c, str);
+		x += ui_measure_character(c);
 	}
 	return x;
 }
