@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 package com.artifex.mupdf.fitz;
 
 public class PDFDocument extends Document
@@ -7,6 +29,7 @@ public class PDFDocument extends Document
 	}
 
 	private static native long newNative();
+	protected native void finalize();
 
 	protected PDFDocument(long p) {
 		super(p);
@@ -46,6 +69,7 @@ public class PDFDocument extends Document
 
 	public native PDFGraftMap newPDFGraftMap();
 	public native PDFObject graftObject(PDFObject obj);
+	public native void graftPage(int pageTo, PDFDocument src, int pageFrom);
 
 	private native PDFObject addStreamBuffer(Buffer buf, Object obj, boolean compressed);
 	private native PDFObject addStreamString(String str, Object obj, boolean compressed);
@@ -100,6 +124,7 @@ public class PDFDocument extends Document
 	public native PDFObject addCJKFont(Font font, int ordering, int wmode, boolean serif);
 	public native PDFObject addFont(Font font);
 	public native boolean hasUnsavedChanges();
+	public native boolean wasRepaired();
 	public native boolean canBeSavedIncrementally();
 
 	public native void save(String filename, String options);
@@ -130,4 +155,20 @@ public class PDFDocument extends Document
 	public native int countUnsavedVersions();
 	public native int validateChangeHistory();
 	public native boolean wasPureXFA();
+	public native boolean wasLinearized();
+
+	public native void enableJournal();
+
+	public native int undoRedoPosition();
+	public native int undoRedoSteps();
+	public native String undoRedoStep(int step);
+
+	public native boolean canUndo();
+	public native boolean canRedo();
+	public native void undo();
+	public native void redo();
+
+	public native void beginOperation(String operation);
+	public native void beginImplicitOperation();
+	public native void endOperation();
 }
