@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "pdfapp.h"
 
 #include <X11/Xlib.h>
@@ -65,8 +87,8 @@ extern void ximage_blit(Drawable d, GC gc, int dstx, int dsty,
 	unsigned char *srcdata,
 	int srcx, int srcy, int srcw, int srch, int srcstride);
 
-void windrawstringxor(pdfapp_t *app, int x, int y, char *s);
-void cleanup(pdfapp_t *app);
+static void windrawstringxor(pdfapp_t *app, int x, int y, char *s);
+static void cleanup(pdfapp_t *app);
 
 static Display *xdpy;
 static Atom XA_CLIPBOARD;
@@ -170,7 +192,7 @@ void winprint(pdfapp_t *app)
 	fprintf(stderr, "The MuPDF library supports printing, but this application currently does not\n");
 }
 
-char *winpassword(pdfapp_t *app, char *filename)
+char *winpassword(pdfapp_t *app, char *fname)
 {
 	char *r = password;
 	password = NULL;
@@ -309,27 +331,28 @@ int wingetsavepath(pdfapp_t *app, char *buf, int len)
 	return 0;
 }
 
-void winreplacefile(char *source, char *target)
+void winreplacefile(pdfapp_t *app, char *source, char *target)
 {
-	rename(source, target);
+	if (rename(source, target) == -1)
+		pdfapp_warn(app, "unable to rename file");
 }
 
-void wincopyfile(char *source, char *target)
+void wincopyfile(pdfapp_t *app, char *source, char *target)
 {
 	FILE *in, *out;
 	char buf[32 << 10];
-	int n;
+	size_t n;
 
 	in = fopen(source, "rb");
 	if (!in)
 	{
-		winerror(&gapp, "cannot open source file for copying");
+		pdfapp_error(app, "cannot open source file for copying");
 		return;
 	}
 	out = fopen(target, "wb");
 	if (!out)
 	{
-		winerror(&gapp, "cannot open target file for copying");
+		pdfapp_error(app, "cannot open target file for copying");
 		fclose(in);
 		return;
 	}
@@ -341,7 +364,7 @@ void wincopyfile(char *source, char *target)
 		if (n < sizeof buf)
 		{
 			if (ferror(in))
-				winerror(&gapp, "cannot read data from source file");
+				pdfapp_error(app, "cannot read data from source file");
 			break;
 		}
 	}
@@ -350,7 +373,7 @@ void wincopyfile(char *source, char *target)
 	fclose(in);
 }
 
-void cleanup(pdfapp_t *app)
+static void cleanup(pdfapp_t *app)
 {
 	fz_context *ctx = app->ctx;
 
@@ -617,7 +640,7 @@ void winadvancetimer(pdfapp_t *app, float duration)
 	advance_scheduled = 1;
 }
 
-void windrawstringxor(pdfapp_t *app, int x, int y, char *s)
+static void windrawstringxor(pdfapp_t *app, int x, int y, char *s)
 {
 	int prevfunction;
 	XGCValues xgcv;
@@ -643,7 +666,7 @@ void windrawstring(pdfapp_t *app, int x, int y, char *s)
 	XDrawString(xdpy, xwin, xgc, x, y, s, strlen(s));
 }
 
-void docopy(pdfapp_t *app, Atom copy_target)
+static void docopy(pdfapp_t *app, Atom copy_target)
 {
 	unsigned short copyucs2[16 * 1024];
 	char *latin1 = copylatin1;
@@ -678,7 +701,7 @@ void windocopy(pdfapp_t *app)
 	docopy(app, XA_PRIMARY);
 }
 
-void onselreq(Window requestor, Atom selection, Atom target, Atom property, Time time)
+static void onselreq(Window requestor, Atom selection, Atom target, Atom property, Time time)
 {
 	XEvent nevt;
 
@@ -773,7 +796,7 @@ void winopenuri(pdfapp_t *app, char *buf)
 			execlp(browser, browser, buf, (char*)0);
 			fprintf(stderr, "cannot exec '%s'\n", browser);
 		}
-		exit(0);
+		_exit(0);
 	}
 	waitpid(pid, NULL, 0);
 }
@@ -783,7 +806,7 @@ int winquery(pdfapp_t *app, const char *query)
 	return QUERY_NO;
 }
 
-int wingetcertpath(char *buf, int len)
+int wingetcertpath(pdfapp_t *app, char *buf, int len)
 {
 	return 0;
 }
@@ -801,11 +824,11 @@ static void onkey(int c, int modifiers)
 	if (!gapp.issearching && c == 'P')
 	{
 		struct timeval now;
-		struct timeval tmo;
-		tmo.tv_sec = 2;
-		tmo.tv_usec = 0;
+		struct timeval t;
+		t.tv_sec = 2;
+		t.tv_usec = 0;
 		gettimeofday(&now, NULL);
-		timeradd(&now, &tmo, &tmo_at);
+		timeradd(&now, &t, &tmo_at);
 		showingpage = 1;
 		winrepaint(&gapp);
 		return;
@@ -1042,7 +1065,7 @@ int main(int argc, char **argv)
 			case ClientMessage:
 				if (xevt.xclient.message_type == WM_RELOAD_PAGE)
 					pdfapp_reloadpage(&gapp);
-				else if (xevt.xclient.format == 32 && xevt.xclient.data.l[0] == WM_DELETE_WINDOW)
+				else if (xevt.xclient.format == 32 && ((Atom) xevt.xclient.data.l[0]) == WM_DELETE_WINDOW)
 					closing = 1;
 				break;
 			}
