@@ -1,5 +1,32 @@
-#include "fitz-imp.h"
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
+#include "mupdf/fitz.h"
+
+#include "context-imp.h"
+#include "color-imp.h"
 #include "draw-imp.h"
+#include "glyph-imp.h"
+#include "pixmap-imp.h"
 
 #include <string.h>
 #include <assert.h>
@@ -18,15 +45,11 @@
 /* Enable the following to help debug graphics stack pushes/pops */
 #undef DUMP_STACK_CHANGES
 
-typedef struct fz_draw_device_s fz_draw_device;
-
 enum {
 	FZ_DRAWDEV_FLAGS_TYPE3 = 1,
 };
 
-typedef struct fz_draw_state_s fz_draw_state;
-
-struct fz_draw_state_s {
+typedef struct {
 	fz_irect scissor;
 	fz_pixmap *dest;
 	fz_pixmap *mask;
@@ -38,9 +61,9 @@ struct fz_draw_state_s {
 	fz_matrix ctm;
 	float xstep, ystep;
 	fz_irect area;
-};
+} fz_draw_state;
 
-struct fz_draw_device_s
+typedef struct fz_draw_device
 {
 	fz_device super;
 	fz_matrix transform;
@@ -55,7 +78,8 @@ struct fz_draw_device_s
 	fz_draw_state *stack;
 	int stack_cap;
 	fz_draw_state init_stack[STACK_SIZE];
-};
+	fz_shade_color_cache *shade_cache;
+} fz_draw_device;
 
 #ifdef DUMP_GROUP_BLENDS
 
@@ -155,7 +179,7 @@ static void grow_stack(fz_context *ctx, fz_draw_device *dev)
 	{
 		stack = fz_realloc_array(ctx, dev->stack, max, fz_draw_state);
 	}
-	dev->stack = stack;
+	dev->stack = Memento_label(stack, "draw_stack");
 	dev->stack_cap = max;
 }
 
@@ -501,10 +525,11 @@ resolve_color(fz_context *ctx,
 	if (color_params.op == 0 || !fz_colorspace_is_subtractive(ctx, dest->colorspace))
 		op = NULL;
 
-	/* Device Gray is additive, but seems to still be counted for overprint
-	 * (see Ghent_V3.0/030_Gray_K_black_OP_x1a.pdf 030.pdf). */
 	else if (devgray)
 	{
+		/* Device Gray is additive, but seems to still be
+		 * counted for overprint (see
+		 * Ghent_V3.0/030_Gray_K_black_OP_x1a.pdf 030.pdf). */
 	}
 
 	/* If we are in a CMYK space (i.e. not a devn one, given we know we are subtractive at this point),
@@ -538,7 +563,7 @@ resolve_color(fz_context *ctx,
 	}
 	colorbv[i] = alpha * 255;
 
-	/* op && !devn => overpinting in cmyk or devicegray. */
+	/* op && !devn => overprinting in cmyk or devicegray. */
 	if (op && !devn)
 	{
 		/* We are overprinting, so protect all spots. */
@@ -574,7 +599,7 @@ push_group_for_separations(fz_context *ctx, fz_draw_device *dev, fz_color_params
 		dcs = oi;
 	}
 
-	/* Not needed if dest has the seps, and we are not using a proof or the target is the same as the prooof and we don't have an oi or the target is the same as the oi */
+	/* Not needed if dest has the seps, and we are not using a proof or the target is the same as the proof and we don't have an oi or the target is the same as the oi */
 	if ((clone == dev->stack[0].dest->seps) && (dev->proof_cs == NULL || dev->proof_cs == dev->stack[0].dest->colorspace) && (oi == NULL || oi == dev->stack[0].dest->colorspace))
 	{
 		fz_drop_separations(ctx, clone);
@@ -622,7 +647,7 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 		flatness = 0.001f;
 
 	bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
-	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, &bbox, &bbox))
+	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, &bbox))
 		return;
 
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
@@ -634,7 +659,7 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 	if (state->shape)
 	{
 		if (!rast->fns.reusable)
-			fz_flatten_fill_path(ctx, rast, path, ctm, flatness, &bbox, NULL);
+			fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, NULL);
 
 		colorbv[0] = 255;
 		fz_convert_rasterizer(ctx, rast, even_odd, state->shape, colorbv, 0);
@@ -642,7 +667,7 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 	if (state->group_alpha)
 	{
 		if (!rast->fns.reusable)
-			fz_flatten_fill_path(ctx, rast, path, ctm, flatness, &bbox, NULL);
+			fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, NULL);
 
 		colorbv[0] = alpha * 255;
 		fz_convert_rasterizer(ctx, rast, even_odd, state->group_alpha, colorbv, 0);
@@ -685,7 +710,7 @@ fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const
 		flatness = 0.001f;
 
 	bbox = fz_intersect_irect(fz_pixmap_bbox_no_ctx(state->dest), state->scissor);
-	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, &bbox, &bbox))
+	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, &bbox))
 		return;
 
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
@@ -706,7 +731,7 @@ fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const
 	if (state->shape)
 	{
 		if (!rast->fns.reusable)
-			(void)fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, &bbox, NULL);
+			(void)fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, NULL);
 
 		colorbv[0] = 255;
 		fz_convert_rasterizer(ctx, rast, 0, state->shape, colorbv, 0);
@@ -714,7 +739,7 @@ fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const
 	if (state->group_alpha)
 	{
 		if (!rast->fns.reusable)
-			(void)fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, &bbox, NULL);
+			(void)fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, NULL);
 
 		colorbv[0] = 255 * alpha;
 		fz_convert_rasterizer(ctx, rast, 0, state->group_alpha, colorbv, 0);
@@ -771,7 +796,7 @@ fz_draw_clip_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 		bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
 	}
 
-	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, &bbox, &bbox) || fz_is_rect_rasterizer(ctx, rast))
+	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, &bbox) || fz_is_rect_rasterizer(ctx, rast))
 	{
 		state[1].scissor = bbox;
 		state[1].mask = NULL;
@@ -788,12 +813,12 @@ fz_draw_clip_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 	if (state[1].shape)
 	{
 		state[1].shape = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].shape);
+		fz_copy_pixmap_rect(ctx, state[1].shape, state[0].shape, bbox, dev->default_cs);
 	}
 	if (state[1].group_alpha)
 	{
 		state[1].group_alpha = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].group_alpha);
+		fz_copy_pixmap_rect(ctx, state[1].group_alpha, state[0].group_alpha, bbox, dev->default_cs);
 	}
 
 	fz_convert_rasterizer(ctx, rast, even_odd, state[1].mask, NULL, 0);
@@ -849,7 +874,7 @@ fz_draw_clip_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path,
 		bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
 	}
 
-	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, &bbox, &bbox))
+	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, &bbox))
 	{
 		state[1].scissor = bbox;
 		state[1].mask = NULL;
@@ -873,12 +898,12 @@ fz_draw_clip_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path,
 	if (state->shape)
 	{
 		state[1].shape = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].shape);
+		fz_copy_pixmap_rect(ctx, state[1].shape, state[0].shape, bbox, dev->default_cs);
 	}
 	if (state->group_alpha)
 	{
 		state[1].group_alpha = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].group_alpha);
+		fz_copy_pixmap_rect(ctx, state[1].group_alpha, state[0].group_alpha, bbox, dev->default_cs);
 	}
 
 	fz_convert_rasterizer(ctx, rast, 0, state[1].mask, NULL, 0);
@@ -918,7 +943,7 @@ draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_glyph *glyph,
 	skip_y = y - glyph->y - yorig;
 
 	msk = glyph->pixmap;
-	dp = dst->samples + (unsigned int)((y - dst->y) * dst->stride + (x - dst->x) * dst->n);
+	dp = dst->samples + (y - dst->y) * (size_t)dst->stride + (x - dst->x) * (size_t)dst->n;
 	if (msk == NULL)
 	{
 		fz_paint_glyph(colorbv, dst, dp, glyph, w, h, skip_x, skip_y, eop);
@@ -1031,7 +1056,8 @@ fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 					fz_matrix mat;
 					mat.a = pixmap->w; mat.b = mat.c = 0; mat.d = pixmap->h;
 					mat.e = x + pixmap->x; mat.f = y + pixmap->y;
-					fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, mat, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, eop);
+					mat = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, mat);
+					fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, mat, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
 				}
 				fz_drop_glyph(ctx, glyph);
 			}
@@ -1104,13 +1130,25 @@ fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const
 			glyph = fz_render_stroked_glyph(ctx, span->font, gid, &trm, ctm, stroke, &state->scissor, aa);
 			if (glyph)
 			{
+				fz_pixmap *pixmap = glyph->pixmap;
 				int x = (int)trm.e;
 				int y = (int)trm.f;
-				draw_glyph(colorbv, state->dest, glyph, x, y, &state->scissor, eop);
-				if (state->shape)
-					draw_glyph(&solid, state->shape, glyph, x, y, &state->scissor, 0);
-				if (state->group_alpha)
-					draw_glyph(&alpha_byte, state->group_alpha, glyph, x, y, &state->scissor, 0);
+				if (pixmap == NULL || pixmap->n == 1)
+				{
+					draw_glyph(colorbv, state->dest, glyph, x, y, &state->scissor, eop);
+					if (state->shape)
+						draw_glyph(&solid, state->shape, glyph, x, y, &state->scissor, 0);
+					if (state->group_alpha)
+						draw_glyph(&alpha_byte, state->group_alpha, glyph, x, y, &state->scissor, 0);
+				}
+				else
+				{
+					fz_matrix mat;
+					mat.a = pixmap->w; mat.b = mat.c = 0; mat.d = pixmap->h;
+					mat.e = x + pixmap->x; mat.f = y + pixmap->y;
+					mat = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, mat);
+					fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, mat, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
+				}
 				fz_drop_glyph(ctx, glyph);
 			}
 			else
@@ -1177,14 +1215,14 @@ fz_draw_clip_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 	if (state->shape)
 	{
 		state[1].shape = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].shape);
+		fz_copy_pixmap_rect(ctx, state[1].shape, state[0].shape, bbox, dev->default_cs);
 	}
 	else
 		state[1].shape = NULL;
 	if (state->group_alpha)
 	{
 		state[1].group_alpha = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, state[1].group_alpha);
+		fz_copy_pixmap_rect(ctx, state[1].group_alpha, state[0].group_alpha, bbox, dev->default_cs);
 	}
 	else
 		state[1].group_alpha = NULL;
@@ -1301,14 +1339,14 @@ fz_draw_clip_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text,
 	if (state->shape)
 	{
 		state[1].shape = shape = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, shape);
+		fz_copy_pixmap_rect(ctx, state[1].shape, state[0].shape, bbox, dev->default_cs);
 	}
 	else
 		shape = state->shape;
 	if (state->group_alpha)
 	{
 		state[1].group_alpha = group_alpha = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
-		fz_clear_pixmap(ctx, group_alpha);
+		fz_copy_pixmap_rect(ctx, state[1].group_alpha, state[0].group_alpha, bbox, dev->default_cs);
 	}
 	else
 		group_alpha = NULL;
@@ -1511,7 +1549,7 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 		else
 			eop = NULL;
 
-		fz_paint_shade(ctx, shade, colorspace, ctm, dest, color_params, bbox, eop);
+		fz_paint_shade(ctx, shade, colorspace, ctm, dest, color_params, bbox, eop, &dev->shade_cache);
 		if (shape)
 			fz_clear_pixmap_rect_with_value(ctx, shape, 255, bbox);
 		if (group_alpha)
@@ -1566,6 +1604,9 @@ fz_transform_pixmap(fz_context *ctx, fz_draw_device *dev, const fz_pixmap *image
 {
 	fz_pixmap *scaled;
 
+	if (clip != NULL && fz_is_empty_irect(*clip))
+		return NULL;
+
 	if (ctm->a != 0 && ctm->b == 0 && ctm->c == 0 && ctm->d != 0)
 	{
 		/* Unrotated or X-flip or Y-flip or XY-flip */
@@ -1663,37 +1704,11 @@ convert_pixmap_for_painting(fz_context *ctx, fz_pixmap *pixmap, fz_colorspace *m
 	return converted;
 }
 
-static void
-fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix in_ctm, float alpha, fz_color_params color_params)
+static fz_irect
+find_src_area_required(fz_matrix local_ctm, fz_image *image, fz_irect clip)
 {
-	fz_draw_device *dev = (fz_draw_device*)devp;
-	fz_matrix local_ctm = fz_concat(in_ctm, dev->transform);
-	fz_pixmap *pixmap;
-	int after;
-	int dx, dy;
-	fz_draw_state *state = &dev->stack[dev->top];
-	fz_colorspace *model;
-	fz_irect clip;
 	fz_matrix inverse;
 	fz_irect src_area;
-	fz_colorspace *src_cs;
-	fz_overprint op = { { 0 } };
-	fz_overprint *eop = &op;
-
-	if (alpha == 0)
-		return;
-
-	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, color_params, dev->default_cs);
-	model = state->dest->colorspace;
-
-	clip = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
-
-	if (image->w == 0 || image->h == 0)
-		return;
-
-	if (color_params.op == 0)
-		eop = NULL;
 
 	/* ctm maps the image (expressed as the unit square) onto the
 	 * destination device. Reverse that to get a mapping from
@@ -1726,10 +1741,48 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 		sane.x1 = image->w;
 		sane.y1 = image->h;
 		src_area = fz_intersect_irect(src_area, sane);
-		if (fz_is_empty_irect(src_area))
-			return;
 	}
 
+	return src_area;
+}
+
+static void
+fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix in_ctm, float alpha, fz_color_params color_params)
+{
+	fz_draw_device *dev = (fz_draw_device*)devp;
+	fz_matrix local_ctm = fz_concat(in_ctm, dev->transform);
+	fz_pixmap *pixmap;
+	int after;
+	int dx, dy;
+	fz_draw_state *state = &dev->stack[dev->top];
+	fz_colorspace *model;
+	fz_irect clip;
+	fz_irect src_area;
+	fz_colorspace *src_cs;
+	fz_overprint op = { { 0 } };
+	fz_overprint *eop = &op;
+
+	if (alpha == 0)
+		return;
+
+	if (dev->top == 0 && dev->resolve_spots)
+		state = push_group_for_separations(ctx, dev, color_params, dev->default_cs);
+	model = state->dest->colorspace;
+
+	clip = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
+
+	if (image->w == 0 || image->h == 0 || fz_is_empty_irect(clip))
+		return;
+
+	if (color_params.op == 0)
+		eop = NULL;
+
+	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+
+	src_area = find_src_area_required(local_ctm, image, clip);
+	if (fz_is_empty_irect(src_area))
+		return;
+
 	pixmap = fz_get_pixmap_from_image(ctx, image, &src_area, &local_ctm, &dx, &dy);
 	src_cs = fz_default_colorspace(ctx, dev->default_cs, pixmap->colorspace);
 
@@ -1798,7 +1851,7 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 				pixmap = convert_pixmap_for_painting(ctx, pixmap, model, src_cs, state->dest, color_params, dev, &eop);
 		}
 
-		fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, eop);
+		fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
 
 		if (state->blendmode & FZ_BLEND_KNOCKOUT)
 			fz_knockout_end(ctx, dev);
@@ -1821,7 +1874,6 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 	int dx, dy;
 	fz_draw_state *state = &dev->stack[dev->top];
 	fz_irect clip;
-	fz_matrix inverse;
 	fz_irect src_area;
 	fz_colorspace *colorspace = NULL;
 	fz_overprint op = { { 0 } };
@@ -1842,40 +1894,11 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 	if (image->w == 0 || image->h == 0)
 		return;
 
-	/* ctm maps the image (expressed as the unit square) onto the
-	 * destination device. Reverse that to get a mapping from
-	 * the destination device to the source pixels. */
-	if (fz_try_invert_matrix(&inverse, local_ctm))
-	{
-		/* Not invertible. Could just bail? Use the whole image
-		 * for now. */
-		src_area.x0 = 0;
-		src_area.x1 = image->w;
-		src_area.y0 = 0;
-		src_area.y1 = image->h;
-	}
-	else
-	{
-		float exp;
-		fz_rect rect;
-		fz_irect sane;
-		/* We want to scale from image coords, not from unit square */
-		inverse = fz_post_scale(inverse, image->w, image->h);
-		/* Are we scaling up or down? exp < 1 means scaling down. */
-		exp = fz_matrix_max_expansion(inverse);
-		rect = fz_rect_from_irect(clip);
-		rect = fz_transform_rect(rect, inverse);
-		/* Allow for support requirements for scalers. */
-		rect = fz_expand_rect(rect, fz_max(exp, 1) * 4);
-		src_area = fz_irect_from_rect(rect);
-		sane.x0 = 0;
-		sane.y0 = 0;
-		sane.x1 = image->w;
-		sane.y1 = image->h;
-		src_area = fz_intersect_irect(src_area, sane);
-		if (fz_is_empty_irect(src_area))
-			return;
-	}
+	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+
+	src_area = find_src_area_required(local_ctm, image, clip);
+	if (fz_is_empty_irect(src_area))
+		return;
 
 	pixmap = fz_get_pixmap_from_image(ctx, image, &src_area, &local_ctm, &dx, &dy);
 
@@ -1907,7 +1930,7 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 
 		eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
 
-		fz_paint_image_with_color(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, colorbv, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, eop);
+		fz_paint_image_with_color(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, colorbv, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
 
 		if (state->blendmode & FZ_BLEND_KNOCKOUT)
 			fz_knockout_end(ctx, dev);
@@ -1930,6 +1953,7 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 	fz_draw_state *state = push_stack(ctx, dev, "clip image mask");
 	fz_colorspace *model = state->dest->colorspace;
 	fz_irect clip;
+	fz_irect src_area;
 
 	fz_var(pixmap);
 
@@ -1949,9 +1973,18 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 		return;
 	}
 
+	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+
+	src_area = find_src_area_required(local_ctm, image, clip);
+	if (fz_is_empty_irect(src_area))
+	{
 #ifdef DUMP_GROUP_BLENDS
-	dump_spaces(dev->top-1, "Clip (image mask) begin\n");
+		dump_spaces(dev->top-1, "Clip (image mask) (empty source area) begin\n");
 #endif
+		state[1].scissor = fz_empty_irect;
+		state[1].mask = NULL;
+		return;
+	}
 
 	bbox = fz_irect_from_rect(fz_transform_rect(fz_unit_rect, local_ctm));
 	bbox = fz_intersect_irect(bbox, state->scissor);
@@ -1960,10 +1993,23 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 		fz_rect tscissor = fz_transform_rect(scissor, dev->transform);
 		bbox = fz_intersect_irect(bbox, fz_irect_from_rect(tscissor));
 	}
+	if (!fz_is_valid_irect(bbox))
+	{
+#ifdef DUMP_GROUP_BLENDS
+		dump_spaces(dev->top-1, "Clip (image mask) (invalid) begin\n");
+#endif
+		state[1].scissor = fz_empty_irect;
+		state[1].mask = NULL;
+		return;
+	}
+
+#ifdef DUMP_GROUP_BLENDS
+	dump_spaces(dev->top-1, "Clip (image mask) begin\n");
+#endif
 
 	fz_try(ctx)
 	{
-		pixmap = fz_get_pixmap_from_image(ctx, image, NULL, &local_ctm, &dx, &dy);
+		pixmap = fz_get_pixmap_from_image(ctx, image, &src_area, &local_ctm, &dx, &dy);
 
 		state[1].mask = fz_new_pixmap_with_bbox(ctx, NULL, bbox, NULL, 1);
 		fz_clear_pixmap(ctx, state[1].mask);
@@ -2013,7 +2059,7 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 			fz_dump_blend(ctx, "/GA=", state[1].group_alpha);
 #endif
 
-		fz_paint_image(ctx, state[1].mask, &bbox, state[1].shape, state[1].group_alpha, pixmap, local_ctm, 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, 0);
+		fz_paint_image(ctx, state[1].mask, &bbox, state[1].shape, state[1].group_alpha, pixmap, local_ctm, 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), 0);
 
 #ifdef DUMP_GROUP_BLENDS
 		fz_dump_blend(ctx, " to get ", state[1].mask);
@@ -2480,7 +2526,7 @@ fz_cmp_tile_key(fz_context *ctx, void *k0_, void *k1_)
 }
 
 static void
-fz_format_tile_key(fz_context *ctx, char *s, int n, void *key_)
+fz_format_tile_key(fz_context *ctx, char *s, size_t n, void *key_)
 {
 	tile_key *key = (tile_key *)key_;
 	fz_snprintf(s, n, "(tile id=%x, ctm=%g %g %g %g, cs=%x, shape=%d, ga=%d)",
@@ -2490,6 +2536,7 @@ fz_format_tile_key(fz_context *ctx, char *s, int n, void *key_)
 
 static const fz_store_type fz_tile_store_type =
 {
+	"struct tile_record",
 	fz_make_hash_tile_key,
 	fz_keep_tile_key,
 	fz_drop_tile_key,
@@ -2722,13 +2769,14 @@ fz_draw_end_tile(fz_context *ctx, fz_device *devp)
 	if (state[0].group_alpha)
 		fz_dump_blend(ctx, "/GA=", state[0].group_alpha);
 #endif
-	dest = fz_new_pixmap_from_pixmap(ctx, state[1].dest, NULL);
-
+	fz_var(dest);
 	fz_var(shape);
 	fz_var(group_alpha);
 
 	fz_try(ctx)
 	{
+		dest = fz_new_pixmap_from_pixmap(ctx, state[1].dest, NULL);
+
 		shape = fz_new_pixmap_from_pixmap(ctx, state[1].shape, NULL);
 		group_alpha = fz_new_pixmap_from_pixmap(ctx, state[1].group_alpha, NULL);
 
@@ -2761,67 +2809,66 @@ fz_draw_end_tile(fz_context *ctx, fz_device *devp)
 				}
 			}
 		}
+
+		/* Now we try to cache the tiles. Any failure here will just result in us not caching. */
+		if (state[1].encache && state[1].id != 0)
+		{
+			tile_record *tile = NULL;
+			tile_key *key = NULL;
+			fz_var(tile);
+			fz_var(key);
+			fz_try(ctx)
+			{
+				tile_record *existing_tile;
+
+				tile = fz_new_tile_record(ctx, state[1].dest, state[1].shape, state[1].group_alpha);
+
+				key = fz_malloc_struct(ctx, tile_key);
+				key->refs = 1;
+				key->id = state[1].id;
+				key->ctm[0] = ctm.a;
+				key->ctm[1] = ctm.b;
+				key->ctm[2] = ctm.c;
+				key->ctm[3] = ctm.d;
+				key->cs = fz_keep_colorspace_store_key(ctx, state[1].dest->colorspace);
+				key->has_shape = (state[1].shape != NULL);
+				key->has_group_alpha = (state[1].group_alpha != NULL);
+				existing_tile = fz_store_item(ctx, key, tile, fz_tile_size(ctx, tile), &fz_tile_store_type);
+				if (existing_tile)
+				{
+					/* We already have a tile. This will either have been
+					 * produced by a racing thread, or there is already
+					 * an entry for this one in the store. */
+					fz_drop_tile_record(ctx, tile);
+					tile = existing_tile;
+				}
+			}
+			fz_always(ctx)
+			{
+				fz_drop_tile_key(ctx, key);
+				fz_drop_tile_record(ctx, tile);
+			}
+			fz_catch(ctx)
+			{
+				/* Do nothing */
+			}
+		}
 	}
 	fz_always(ctx)
 	{
 		fz_drop_pixmap(ctx, dest);
 		fz_drop_pixmap(ctx, shape);
 		fz_drop_pixmap(ctx, group_alpha);
+		fz_drop_pixmap(ctx, state[1].dest);
+		state[1].dest = NULL;
+		fz_drop_pixmap(ctx, state[1].shape);
+		state[1].shape = NULL;
+		fz_drop_pixmap(ctx, state[1].group_alpha);
+		state[1].group_alpha = NULL;
 	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
-	/* Now we try to cache the tiles. Any failure here will just result in us not caching. */
-	if (state[1].encache && state[1].id != 0)
-	{
-		tile_record *tile = NULL;
-		tile_key *key = NULL;
-		fz_var(tile);
-		fz_var(key);
-		fz_try(ctx)
-		{
-			tile_record *existing_tile;
-
-			tile = fz_new_tile_record(ctx, state[1].dest, state[1].shape, state[1].group_alpha);
-
-			key = fz_malloc_struct(ctx, tile_key);
-			key->refs = 1;
-			key->id = state[1].id;
-			key->ctm[0] = ctm.a;
-			key->ctm[1] = ctm.b;
-			key->ctm[2] = ctm.c;
-			key->ctm[3] = ctm.d;
-			key->cs = fz_keep_colorspace_store_key(ctx, state[1].dest->colorspace);
-			key->has_shape = (state[1].shape != NULL);
-			key->has_group_alpha = (state[1].group_alpha != NULL);
-			existing_tile = fz_store_item(ctx, key, tile, fz_tile_size(ctx, tile), &fz_tile_store_type);
-			if (existing_tile)
-			{
-				/* We already have a tile. This will either have been
-				 * produced by a racing thread, or there is already
-				 * an entry for this one in the store. */
-				fz_drop_tile_record(ctx, tile);
-				tile = existing_tile;
-			}
-		}
-		fz_always(ctx)
-		{
-			fz_drop_tile_key(ctx, key);
-			fz_drop_tile_record(ctx, tile);
-		}
-		fz_catch(ctx)
-		{
-			/* Do nothing */
-		}
-	}
-
-	fz_drop_pixmap(ctx, state[1].dest);
-	state[1].dest = NULL;
-	fz_drop_pixmap(ctx, state[1].shape);
-	state[1].shape = NULL;
-	fz_drop_pixmap(ctx, state[1].group_alpha);
-	state[1].group_alpha = NULL;
-
 #ifdef DUMP_GROUP_BLENDS
 	fz_dump_blend(ctx, " to get ", state[0].dest);
 	if (state[0].shape)
@@ -2912,9 +2959,10 @@ fz_draw_drop_device(fz_context *ctx, fz_device *devp)
 	fz_drop_scale_cache(ctx, dev->cache_x);
 	fz_drop_scale_cache(ctx, dev->cache_y);
 	fz_drop_rasterizer(ctx, rast);
+	fz_drop_shade_color_cache(ctx, dev->shade_cache);
 }
 
-fz_device *
+static fz_device *
 new_draw_device(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_aa_context *aa, const fz_irect *clip, fz_colorspace *proof_cs)
 {
 	fz_draw_device *dev = fz_new_derived_device(ctx, fz_draw_device);
@@ -3019,79 +3067,24 @@ new_draw_device(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_
 	return (fz_device*)dev;
 }
 
-/*
-	Create a device to draw on a pixmap.
-
-	dest: Target pixmap for the draw device. See fz_new_pixmap*
-	for how to obtain a pixmap. The pixmap is not cleared by the
-	draw device, see fz_clear_pixmap* for how to clear it prior to
-	calling fz_new_draw_device. Free the device by calling
-	fz_drop_device.
-
-	transform: Transform from user space in points to device space in pixels.
-*/
 fz_device *
 fz_new_draw_device(fz_context *ctx, fz_matrix transform, fz_pixmap *dest)
 {
 	return new_draw_device(ctx, transform, dest, NULL, NULL, NULL);
 }
 
-/*
-	Create a device to draw on a pixmap.
-
-	dest: Target pixmap for the draw device. See fz_new_pixmap*
-	for how to obtain a pixmap. The pixmap is not cleared by the
-	draw device, see fz_clear_pixmap* for how to clear it prior to
-	calling fz_new_draw_device. Free the device by calling
-	fz_drop_device.
-
-	transform: Transform from user space in points to device space in pixels.
-
-	clip: Bounding box to restrict any marking operations of the
-	draw device.
-*/
 fz_device *
 fz_new_draw_device_with_bbox(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_irect *clip)
 {
 	return new_draw_device(ctx, transform, dest, NULL, clip, NULL);
 }
 
-/*
-	Create a device to draw on a pixmap.
-
-	dest: Target pixmap for the draw device. See fz_new_pixmap*
-	for how to obtain a pixmap. The pixmap is not cleared by the
-	draw device, see fz_clear_pixmap* for how to clear it prior to
-	calling fz_new_draw_device. Free the device by calling
-	fz_drop_device.
-
-	transform: Transform from user space in points to device space in pixels.
-
-	proof_cs: Intermediate color space to map though when mapping to
-	color space defined by pixmap.
-*/
 fz_device *
 fz_new_draw_device_with_proof(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, fz_colorspace *cs)
 {
 	return new_draw_device(ctx, transform, dest, NULL, NULL, cs);
 }
 
-/*
-	Create a device to draw on a pixmap.
-
-	dest: Target pixmap for the draw device. See fz_new_pixmap*
-	for how to obtain a pixmap. The pixmap is not cleared by the
-	draw device, see fz_clear_pixmap* for how to clear it prior to
-	calling fz_new_draw_device. Free the device by calling
-	fz_drop_device.
-
-	transform: Transform from user space in points to device space in pixels.
-
-	clip: Bounding box to restrict any marking operations of the
-	draw device.
-
-	proof_cs: Color space to render to prior to mapping to color space defined by pixmap.
-*/
 fz_device *
 fz_new_draw_device_with_bbox_proof(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_irect *clip, fz_colorspace *cs)
 {
@@ -3107,7 +3100,7 @@ fz_new_draw_device_type3(fz_context *ctx, fz_matrix transform, fz_pixmap *dest)
 }
 
 fz_irect *
-fz_bound_path_accurate(fz_context *ctx, fz_irect *bbox, const fz_irect *scissor, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm, float flatness, float linewidth)
+fz_bound_path_accurate(fz_context *ctx, fz_irect *bbox, fz_irect scissor, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm, float flatness, float linewidth)
 {
 	fz_rasterizer *rast = fz_new_rasterizer(ctx, NULL);
 
@@ -3154,9 +3147,6 @@ static int parse_aa_opts(const char *val)
 	return 8;
 }
 
-/*
-	Parse draw device options from a comma separated key-value string.
-*/
 fz_draw_options *
 fz_parse_draw_options(fz_context *ctx, fz_draw_options *opts, const char *args)
 {
@@ -3213,14 +3203,6 @@ fz_parse_draw_options(fz_context *ctx, fz_draw_options *opts, const char *args)
 	return opts;
 }
 
-/*
-
-	Create a new pixmap and draw device, using the specified options.
-
-	options: Options to configure the draw device, and choose the resolution and colorspace.
-	mediabox: The bounds of the page in points.
-	pixmap: An out parameter containing the newly created pixmap.
-*/
 fz_device *
 fz_new_draw_device_with_options(fz_context *ctx, const fz_draw_options *opts, fz_rect mediabox, fz_pixmap **pixmap)
 {
