@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include "pixmap-imp.h"
@@ -179,7 +201,7 @@ png_deinterlace(fz_context *ctx, struct info *info, unsigned int *passw, unsigne
 {
 	unsigned int n = info->n;
 	unsigned int depth = info->depth;
-	unsigned int stride = (info->width * n * depth + 7) / 8;
+	size_t stride = ((size_t)info->width * n * depth + 7) / 8;
 	unsigned char *output;
 	unsigned int p, x, y, k;
 
@@ -427,7 +449,7 @@ png_read_image(fz_context *ctx, struct info *info, const unsigned char *p, size_
 	/* Read IHDR chunk (must come first) */
 
 	size = getuint(p);
-	if (total < 12 || size > total - 12)
+	if (size > total - 12)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in png image");
 
 	if (!memcmp(p + 4, "IHDR", 4))
@@ -559,8 +581,8 @@ png_expand_palette(fz_context *ctx, struct info *info, fz_pixmap *src)
 	unsigned char *sp = src->samples;
 	unsigned char *dp = dst->samples;
 	unsigned int x, y;
-	int dstride = dst->stride - dst->w * dst->n;
-	int sstride = src->stride - src->w * src->n;
+	size_t dstride = dst->stride - dst->w * (size_t)dst->n;
+	size_t sstride = src->stride - src->w * (size_t)src->n;
 
 	dst->xres = src->xres;
 	dst->yres = src->yres;
