@@ -1,4 +1,4 @@
-/* Copyright (C) 2009-2018 Artifex Software, Inc.
+/* Copyright (C) 2009-2020 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -39,7 +39,11 @@
 #include "memory_.h"
 int atexit(void (*)(void));
 #else
+#ifdef MEMENTO_MUPDF_HACKS
 #include "mupdf/memento.h"
+#else
+#include "memento.h"
+#endif
 #include <stdio.h>
 #endif
 #ifndef _MSC_VER
@@ -48,6 +52,7 @@ int atexit(void (*)(void));
 #include <unistd.h>
 #endif
 
+#include <errno.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
@@ -57,6 +62,19 @@ int atexit(void (*)(void));
 #include <stdio.h>
 #endif
 
+/* Workaround VS2012 (and earlier) missing va_copy. */
+#ifdef _MSC_VER
+# if _MSC_VER < 1800 /* Prior to 2013 */
+#  ifndef va_copy
+#   ifdef __va_copy
+#    define va_copy(dst,src) __va_copy(dst,src)
+#   else
+#    define va_copy(dst,src) memcpy(&dst, &src, sizeof(va_list))
+#   endif /* __va_copy */
+#  endif /* va_copy */
+# endif
+#endif
+
 /* Hacks to portably print large sizes */
 #ifdef _MSC_VER
 #define FMTZ "%llu"
@@ -189,7 +207,7 @@ windows_fprintf(FILE *file, const char *fmt, ...)
 #endif
 #endif
 
-#if defined(__linux__)
+#if defined(__linux__) || defined(__OpenBSD__)
 #define MEMENTO_HAS_FORK
 #elif defined(__APPLE__) && defined(__MACH__)
 #define MEMENTO_HAS_FORK
@@ -270,7 +288,10 @@ enum {
     Memento_EventType_deleteArray = 7,
     Memento_EventType_takeRef = 8,
     Memento_EventType_dropRef = 9,
-    Memento_EventType_reference = 10
+    Memento_EventType_reference = 10,
+    Memento_EventType_strdup = 11,
+    Memento_EventType_asprintf = 12,
+    Memento_EventType_vasprintf = 13
 };
 
 static const char *eventType[] =
@@ -285,7 +306,10 @@ static const char *eventType[] =
     "delete[]",
     "takeRef",
     "dropRef",
-    "reference"
+    "reference",
+    "strdup",
+    "asprintf",
+    "vasprintf"
 };
 
 /* When we list leaked blocks at the end of execution, we search for pointers
@@ -402,6 +426,11 @@ static void Memento_initMutex(Memento_mutex *m)
 #endif
 #endif
 
+typedef struct {
+    int begin;
+    int end;
+} Memento_range;
+
 /* And our global structure */
 static struct {
     int            inited;
@@ -424,6 +453,9 @@ static struct {
     int            nextPattern;
     int            patternBit;
     int            leaking;
+    int            hideMultipleReallocs;
+    int            abortOnLeak;
+    int            abortOnCorruption;
     size_t         maxMemory;
     size_t         alloc;
     size_t         peakAlloc;
@@ -432,6 +464,9 @@ static struct {
     size_t         numFrees;
     size_t         numReallocs;
     Memento_mutex  mutex;
+    Memento_range *squeezes;
+    int            squeezes_num;
+    int            squeezes_pos;
 } memento;
 
 #define MEMENTO_EXTRASIZE (sizeof(Memento_BlkHeader) + Memento_PostSize)
@@ -500,10 +535,16 @@ static void *current_addr;
 
 static void error2_cb(void *data, const char *msg, int errnum)
 {
+    (void)data;
+    (void)msg;
+    (void)errnum;
 }
 
 static void syminfo_cb(void *data, uintptr_t pc, const char *symname, uintptr_t symval, uintptr_t symsize)
 {
+    (void)data;
+    (void)symval;
+    (void)symsize;
     if (sizeof(void *) == 4)
         fprintf(stderr, "    0x%08lx %s\n", pc, symname?symname:"?");
     else
@@ -512,6 +553,9 @@ static void syminfo_cb(void *data, uintptr_t pc, const char *symname, uintptr_t
 
 static void error_cb(void *data, const char *msg, int errnum)
 {
+    (void)data;
+    (void)msg;
+    (void)errnum;
     backtrace_syminfo(my_backtrace_state,
                      (uintptr_t)current_addr,
                      syminfo_cb,
@@ -521,6 +565,7 @@ static void error_cb(void *data, const char *msg, int errnum)
 
 static int full_cb(void *data, uintptr_t pc, const char *fname, int line, const char *fn)
 {
+    (void)data;
     if (sizeof(void *) == 4)
         fprintf(stderr, "    0x%08lx %s(%s:%d)\n", pc, fn?fn:"?", fname?fname:"?", line);
     else
@@ -934,8 +979,8 @@ static void Memento_storeDetails(Memento_BlkHeader *head, int type)
     if (count)
         memcpy(&details->stack, &stack[skip], count * sizeof(void *));
 
-    details->type = type;
-    details->count = count;
+    details->type = (char)type;
+    details->count = (char)count;
     details->sequence = memento.sequence;
     details->next = NULL;
     VALGRIND_MAKE_MEM_DEFINED(&head->details_tail, sizeof(head->details_tail));
@@ -1282,6 +1327,7 @@ static int Memento_appBlock(Memento_Blocks    *blks,
                             Memento_BlkHeader *b)
 {
     int result;
+    (void)blks;
     VALGRIND_MAKE_MEM_DEFINED(b, sizeof(Memento_BlkHeader));
     VALGRIND_MAKE_MEM_DEFINED(MEMBLK_TOBLK(b),
                               b->rawsize + Memento_PostSize);
@@ -1412,9 +1458,15 @@ int Memento_listBlocksNested(void)
     /* Now, calculate tree */
     for (b = memento.used.head; b; b = b->next) {
         char *p = MEMBLK_TOBLK(b);
-        int end = (b->rawsize < MEMENTO_PTRSEARCH ? b->rawsize : MEMENTO_PTRSEARCH);
-        for (i = MEMENTO_SEARCH_SKIP; i < end; i += sizeof(void *)) {
-            void *q = *(void **)(&p[i]);
+        size_t end = (b->rawsize < MEMENTO_PTRSEARCH ? b->rawsize : MEMENTO_PTRSEARCH);
+        size_t z;
+        VALGRIND_MAKE_MEM_DEFINED(p, end);
+        if (end > sizeof(void *)-1)
+            end -= sizeof(void *)-1;
+        else
+            end = 0;
+        for (z = MEMENTO_SEARCH_SKIP; z < end; z += sizeof(void *)) {
+            void *q = *(void **)(&p[z]);
             void **r;
 
             /* Do trivial checks on pointer */
@@ -1437,7 +1489,7 @@ int Memento_listBlocksNested(void)
 
                 /* Not interested in pointers to ourself! */
                 if (child == b)
-                        continue;
+                    continue;
 
                 /* We're also assuming acyclicness here. If this is one of
                  * our parents, ignore it. */
@@ -1542,18 +1594,24 @@ static int showInfo(Memento_BlkHeader *b, void *arg)
 {
     Memento_BlkDetails *details;
 
+    (void)arg;
+
     fprintf(stderr, FMTP":(size="FMTZ",num=%d)",
             MEMBLK_TOBLK(b), (FMTZ_CAST)b->rawsize, b->sequence);
     if (b->label)
         fprintf(stderr, " (%s)", b->label);
     fprintf(stderr, "\nEvents:\n");
 
-    details = b->details;
-    while (details)
+    for (details = b->details; details; details = details->next)
     {
+        if (memento.hideMultipleReallocs &&
+            details->type == Memento_EventType_realloc &&
+            details->next &&
+            details->next->type == Memento_EventType_realloc) {
+            continue;
+        }
         fprintf(stderr, "  Event %d (%s)\n", details->sequence, eventType[(int)details->type]);
         Memento_showStacktrace(details->stack, details->count);
-        details = details->next;
     }
     return 0;
 }
@@ -1569,6 +1627,25 @@ void Memento_listBlockInfo(void)
 #endif
 }
 
+#ifdef MEMENTO_DETAILS
+static int
+showBlockInfo(Memento_BlkHeader *b, void *arg)
+{
+    if (arg < MEMBLK_TOBLK(b) || (void *)MEMBLK_POSTPTR(b) <= arg)
+        return 0;
+    return showInfo(b, NULL);
+}
+#endif
+
+void Memento_blockInfo(void *p)
+{
+#ifdef MEMENTO_DETAILS
+    MEMENTO_LOCK();
+    Memento_appBlocks(&memento.used, showBlockInfo, p);
+    MEMENTO_UNLOCK();
+#endif
+}
+
 static int Memento_nonLeakBlocksLeaked(void)
 {
     Memento_BlkHeader *blk = memento.used.head;
@@ -1620,6 +1697,182 @@ void Memento_fin(void)
         fprintf(stderr, "MEMENTO_NEXTFAILAT=%d\n", memento.nextFailAt);
         fprintf(stderr, "MEMENTO_NEXTPATTERN=%d\n", memento.nextPattern);
     }
+    if (Memento_nonLeakBlocksLeaked() && memento.abortOnLeak) {
+        fprintf(stderr, "Calling abort() because blocks were leaked and MEMENTO_ABORT_ON_LEAK is set.\n");
+        abort();
+    }
+}
+
+/* Reads number from <text> using strtol().
+ *
+ * Params:
+ *     text:
+ *         text to read.
+ *     out:
+ *         pointer to output value.
+ *     relative:
+ *         *relative set to 1 if <text> starts with '+' or '-', else set to 0.
+ *     end:
+ *         *end is set to point to next unread character after number.
+ *
+ * Returns 0 on success, else -1.
+ */
+static int read_number(const char *text, int *out, int *relative, char **end)
+{
+    if (text[0] == '+' || text[0] == '-')
+        *relative = 1;
+    else
+        *relative = 0;
+    errno = 0;
+    *out = (int)strtol(text, end, 0 /*base*/);
+    if (errno || *end == text)
+    {
+        fprintf(stderr, "Failed to parse number at start of '%s'.\n", text);
+        return -1;
+    }
+    if (0)
+         fprintf(stderr, "text='%s': *out=%i *relative=%i\n",
+                 text, *out, *relative);
+    return 0;
+}
+
+/* Reads number plus optional delta value from <text>.
+ *
+ * Evaluates <number> or <number>[+|-<delta>]. E.g. text='1234+2' sets *out=1236,
+ * text='1234-1' sets *out=1233.
+ *
+ * Params:
+ *     text:
+ *         text to read.
+ *     out:
+ *         pointer to output value.
+ *     end:
+ *         *end is set to point to next unread character after number.
+ *
+ * Returns 0 on success, else -1.
+ */
+static int read_number_delta(const char *text, int *out, char **end)
+{
+    int e;
+    int relative;
+
+    e = read_number(text, out, &relative, end);
+    if (e)
+        return e;
+    if (relative) {
+        fprintf(stderr, "Base number should not start with '+' or '-' at start of '%s'.\n",
+                text);
+        return -1;
+    }
+    if (*end) {
+        if (**end == '-' || **end == '+') {
+            int delta;
+            e = read_number(*end, &delta, &relative, end);
+            if (e)
+                return e;
+            *out += delta;
+        }
+    }
+    if (0) fprintf(stderr, "text='%s': *out=%i\n", text, *out);
+
+    return 0;
+}
+
+/* Reads range.
+ *
+ * E.g.:
+ *     text='115867-2' sets *begin=115865 *end=115866.
+ *     text='115867-1..+3' sets *begin=115866 *end=115869.
+ *
+ * Supported patterns for text:
+ *     <range>
+ *         <value>             - returns *begin=value *end=*begin+1.
+ *         <value1>..<value2>  - returns *begin=value1 *end=value2.
+ *         <value>..+<number>  - returns *begin=value *end=*begin+number.
+ *     <value>
+ *         <number>
+ *         <number>+<number>
+ *         <number>-<number>
+ *
+ *     <number>: [0-9]+
+ *
+ * If not specified, *end defaults to *begin+1.
+ *
+ * Returns 0 on success, else -1, with *string_end pointing to first unused
+ * character.
+ */
+static int read_number_range(const char *text, int *begin, int *end, char **string_end)
+{
+    int e;
+    e = read_number_delta(text, begin, string_end);
+    if (e)
+        return e;
+    if (string_end && (*string_end)[0] == '.' && (*string_end)[1] == '.') {
+        int relative;
+        e = read_number((*string_end) + 2, end, &relative, string_end);
+        if (e)
+            return e;
+        if (relative)
+            *end += *begin;
+    } else {
+        *end = *begin + 1;
+    }
+    if (*end < *begin) {
+        fprintf(stderr, "Range %i..%i has negative extent, at start of '%s'.\n",
+                *begin, *end, text);
+        return -1;
+    }
+    if (0) fprintf(stderr, "text='%s': *begin=%i *end=%i\n", text, *begin, *end);
+
+    return 0;
+}
+
+/* Format: <range>[,<range>]+
+ *
+ * For description of <range>, see read_number_range() above.
+ *
+ * E.g.:
+ *     MEMENTO_SQUEEZES=1234-2..+4,2345,2350..+2
+ */
+static int Memento_add_squeezes(const char *text)
+{
+    int e = 0;
+    for(;;) {
+        int     begin;
+        int     end;
+        char   *string_end;
+        if (!*text)
+            break;
+        e = read_number_range(text, &begin, &end, &string_end);
+        if (e)
+            break;
+        if (*string_end && *string_end != ',') {
+            fprintf(stderr, "Expecting comma at start of '%s'.\n", string_end);
+            e = -1;
+            break;
+        }
+        fprintf(stderr, "Adding squeeze range %i..%i.\n",
+                begin, end);
+        memento.squeezes_num += 1;
+        memento.squeezes = MEMENTO_UNDERLYING_REALLOC(
+                memento.squeezes,
+                memento.squeezes_num * sizeof(*memento.squeezes)
+                );
+        if (!memento.squeezes) {
+            fprintf(stderr, "Failed to allocate memory for memento.squeezes_num=%i\n",
+                    memento.squeezes_num);
+            e = -1;
+            break;
+        }
+        memento.squeezes[memento.squeezes_num-1].begin = begin;
+        memento.squeezes[memento.squeezes_num-1].end = end;
+
+        if (*string_end == 0)
+            break;
+        text = string_end + 1;
+    }
+
+    return e;
 }
 
 static void Memento_init(void)
@@ -1633,6 +1886,9 @@ static void Memento_init(void)
     memento.free.tail = NULL;
     memento.sequence  = 0;
     memento.countdown = 1024;
+    memento.squeezes  = NULL;
+    memento.squeezes_num = 0;
+    memento.squeezes_pos = 0;
 
     env = getenv("MEMENTO_FAILAT");
     memento.failAt = (env ? atoi(env) : 0);
@@ -1654,6 +1910,26 @@ static void Memento_init(void)
     env = getenv("MEMENTO_PATTERN");
     memento.pattern = (env ? atoi(env) : 0);
 
+    env = getenv("MEMENTO_HIDE_MULTIPLE_REALLOCS");
+    memento.hideMultipleReallocs = (env ? atoi(env) : 0);
+
+    env = getenv("MEMENTO_ABORT_ON_LEAK");
+    memento.abortOnLeak = (env ? atoi(env) : 0);
+
+    env = getenv("MEMENTO_ABORT_ON_CORRUPTION");
+    memento.abortOnCorruption = (env ? atoi(env) : 0);
+
+    env = getenv("MEMENTO_SQUEEZES");
+    if (env) {
+        int e;
+        fprintf(stderr, "Parsing squeeze ranges in MEMENTO_SQUEEZES=%s\n", env);
+        e = Memento_add_squeezes(env);
+        if (e) {
+            fprintf(stderr, "Failed to parse MEMENTO_SQUEEZES=%s\n", env);
+            exit(1);
+        }
+    }
+
     env = getenv("MEMENTO_MAXMEMORY");
     memento.maxMemory = (env ? atoi(env) : 0);
 
@@ -1774,7 +2050,9 @@ static int squeeze(void)
     for (i = 0; i < OPEN_MAX; i++) {
         if (stashed_map[i] == 0) {
             int j = dup(i);
-            stashed_map[j] = i+1;
+            if (j >= 0) {
+                stashed_map[j] = i+1;
+            }
         }
     }
 
@@ -1784,12 +2062,26 @@ static int squeeze(void)
     if (pid == 0) {
         /* Child */
         signal(SIGSEGV, Memento_signal);
+        /* Close the dup-licated fds to avoid them getting corrupted by faulty
+         * code. */
+        for (i = 0; i < OPEN_MAX; i++) {
+            if (stashed_map[i] != 0) {
+                /* We close duplicated fds, just in case child has some bad
+                 * code that modifies/closes random fds. */
+                close(i);
+            }
+        }
         /* In the child, we always fail the next allocation. */
         if (memento.patternBit == 0) {
             memento.patternBit = 1;
         } else
             memento.patternBit <<= 1;
         memento.squeezing = 1;
+
+        /* This is necessary to allow Memento_failThisEventLocked() near the
+         * end to do 'return squeeze();'. */
+        memento.squeezes_num = 0;
+
         return 1;
     }
 
@@ -1804,7 +2096,7 @@ static int squeeze(void)
         int timeout = 30 * 1000 * 1000; /* time out in microseconds! */
         while (waitpid(pid, &status, WNOHANG) == 0) {
             nanosleep(&tm, NULL);
-            timeout -= (tm.tv_nsec/1000);
+            timeout -= (int)(tm.tv_nsec/1000);
             tm.tv_nsec *= 2;
             if (tm.tv_nsec > 999999999)
                 tm.tv_nsec = 999999999;
@@ -1861,6 +2153,7 @@ static void Memento_startFailing(void)
 {
     if (!memento.failing) {
         fprintf(stderr, "Starting to fail...\n");
+        Memento_bt();
         fflush(stderr);
         memento.failing = 1;
         memento.failAt = memento.sequence;
@@ -1977,11 +2270,37 @@ static int Memento_failThisEventLocked(void)
 
     if (Memento_event()) Memento_breakpointLocked();
 
+    if (!memento.squeezing && memento.squeezes_num) {
+        /* Move to next relevant squeeze region if appropriate. */
+        for ( ; memento.squeezes_pos != memento.squeezes_num; memento.squeezes_pos++) {
+            if (memento.sequence < memento.squeezes[memento.squeezes_pos].end)
+                break;
+        }
+
+        /* See whether memento.sequence is within this squeeze region. */
+        if (memento.squeezes_pos < memento.squeezes_num) {
+            int begin = memento.squeezes[memento.squeezes_pos].begin;
+            int end   = memento.squeezes[memento.squeezes_pos].end;
+            if (memento.sequence >= begin && memento.sequence < end) {
+                if (1) {
+                    fprintf(stderr,
+                            "squeezes match memento.sequence=%i: memento.squeezes_pos=%i/%i %i..%i\n",
+                            memento.sequence,
+                            memento.squeezes_pos,
+                            memento.squeezes_num,
+                            memento.squeezes[memento.squeezes_pos].begin,
+                            memento.squeezes[memento.squeezes_pos].end
+                            );
+                }
+                return squeeze();
+            }
+        }
+    }
+
     if ((memento.sequence >= memento.failAt) && (memento.failAt != 0))
         Memento_startFailing();
-    if ((memento.sequence >= memento.squeezeAt) && (memento.squeezeAt != 0)) {
+    if ((memento.squeezes_num==0) && (memento.sequence >= memento.squeezeAt) && (memento.squeezeAt != 0))
         return squeeze();
-    }
 
     if (!memento.failing)
         return 0;
@@ -2021,16 +2340,22 @@ static void *do_malloc(size_t s, int eventType)
     Memento_BlkHeader *memblk;
     size_t             smem = MEMBLK_SIZE(s);
 
-    if (Memento_failThisEventLocked())
+    (void)eventType;
+
+    if (Memento_failThisEventLocked()) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     if (s == 0)
         return NULL;
 
     memento.numMallocs++;
 
-    if (memento.maxMemory != 0 && memento.alloc + s > memento.maxMemory)
+    if (memento.maxMemory != 0 && memento.alloc + s > memento.maxMemory) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     memblk = MEMENTO_UNDERLYING_MALLOC(smem);
     if (memblk == NULL)
@@ -2053,7 +2378,7 @@ static void *do_malloc(size_t s, int eventType)
 #ifdef MEMENTO_DETAILS
     memblk->details       = NULL;
     memblk->details_tail  = &memblk->details;
-    Memento_storeDetails(memblk, Memento_EventType_malloc);
+    Memento_storeDetails(memblk, eventType);
 #endif /* MEMENTO_DETAILS */
     Memento_addBlockHead(&memento.used, memblk, 0);
 
@@ -2063,6 +2388,81 @@ static void *do_malloc(size_t s, int eventType)
     return MEMBLK_TOBLK(memblk);
 }
 
+char *Memento_strdup(const char *text)
+{
+    size_t len = strlen(text) + 1;
+    char *ret;
+
+    if (!memento.inited)
+        Memento_init();
+
+    MEMENTO_LOCK();
+    ret = do_malloc(len, Memento_EventType_strdup);
+    MEMENTO_UNLOCK();
+
+    if (ret != NULL)
+        memcpy(ret, text, len);
+
+    return ret;
+}
+
+int Memento_asprintf(char **ret, const char *format, ...)
+{
+    va_list va;
+    int n;
+    int n2;
+
+    if (!memento.inited)
+        Memento_init();
+
+    va_start(va, format);
+    n = vsnprintf(NULL, 0, format, va);
+    va_end(va);
+    if (n < 0)
+        return n;
+
+    MEMENTO_LOCK();
+    *ret = do_malloc(n+1, Memento_EventType_asprintf);
+    MEMENTO_UNLOCK();
+    if (*ret == NULL)
+        return -1;
+
+    va_start(va, format);
+    n2 = vsnprintf(*ret, n + 1, format, va);
+    va_end(va);
+
+    return n2;
+}
+
+int Memento_vasprintf(char **ret, const char *format, va_list ap)
+{
+    int n;
+    va_list ap2;
+    va_copy(ap2, ap);
+
+    if (!memento.inited)
+        Memento_init();
+
+    n = vsnprintf(NULL, 0, format, ap);
+    if (n < 0) {
+        va_end(ap2);
+        return n;
+    }
+
+    MEMENTO_LOCK();
+    *ret = do_malloc(n+1, Memento_EventType_vasprintf);
+    MEMENTO_UNLOCK();
+    if (*ret == NULL) {
+        va_end(ap2);
+        return -1;
+    }
+
+    n = vsnprintf(*ret, n + 1, format, ap2);
+    va_end(ap2);
+
+    return n;
+}
+
 void *Memento_malloc(size_t s)
 {
     void *ret;
@@ -2073,6 +2473,7 @@ void *Memento_malloc(size_t s)
     MEMENTO_LOCK();
     ret = do_malloc(s, Memento_EventType_malloc);
     MEMENTO_UNLOCK();
+
     return ret;
 }
 
@@ -2085,9 +2486,10 @@ void *Memento_calloc(size_t n, size_t s)
 
     MEMENTO_LOCK();
     block = do_malloc(n*s, Memento_EventType_calloc);
+    MEMENTO_UNLOCK();
     if (block)
         memset(block, 0, n*s);
-    MEMENTO_UNLOCK();
+
     return block;
 }
 
@@ -2128,7 +2530,7 @@ int Memento_checkBytePointerOrNull(void *blk)
         return 0;
     Memento_checkPointerOrNull(blk);
 
-    i = *(unsigned int *)blk;
+    i = *(unsigned char *)blk;
 
     if (i == MEMENTO_PREFILL_UBYTE)
         fprintf(stderr, "Prefill value found - buffer underrun?\n");
@@ -2470,6 +2872,8 @@ static void do_free(void *blk, int eventType)
 {
     Memento_BlkHeader *memblk;
 
+    (void)eventType;
+
     if (Memento_event()) Memento_breakpointLocked();
 
     if (blk == NULL)
@@ -2478,10 +2882,16 @@ static void do_free(void *blk, int eventType)
     memblk = MEMBLK_FROMBLK(blk);
     VALGRIND_MAKE_MEM_DEFINED(memblk, sizeof(*memblk));
     if (checkBlock(memblk, "free"))
+    {
+        if (memento.abortOnCorruption) {
+            fprintf(stderr, "*** memblk corrupted, calling abort()\n");
+            abort();
+        }
         return;
+    }
 
 #ifdef MEMENTO_DETAILS
-    Memento_storeDetails(memblk, Memento_EventType_free);
+    Memento_storeDetails(memblk, eventType);
 #endif
 
     VALGRIND_MAKE_MEM_DEFINED(memblk, sizeof(*memblk));
@@ -2524,13 +2934,17 @@ static void *do_realloc(void *blk, size_t newsize, int type)
     size_t             newsizemem;
     int                flags;
 
-    if (Memento_failThisEventLocked())
+    if (Memento_failThisEventLocked()) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     memblk     = MEMBLK_FROMBLK(blk);
     VALGRIND_MAKE_MEM_DEFINED(memblk, sizeof(*memblk));
-    if (checkBlock(memblk, "realloc"))
+    if (checkBlock(memblk, "realloc")) {
+        errno = ENOMEM;
         return NULL;
+    }
 
 #ifdef MEMENTO_DETAILS
     Memento_storeDetails(memblk, type);
@@ -2541,8 +2955,10 @@ static void *do_realloc(void *blk, size_t newsize, int type)
         Memento_breakpointLocked();
 
     VALGRIND_MAKE_MEM_DEFINED(memblk, sizeof(*memblk));
-    if (memento.maxMemory != 0 && memento.alloc - memblk->rawsize + newsize > memento.maxMemory)
+    if (memento.maxMemory != 0 && memento.alloc - memblk->rawsize + newsize > memento.maxMemory) {
+        errno = ENOMEM;
         return NULL;
+    }
 
     newsizemem = MEMBLK_SIZE(newsize);
     Memento_removeBlock(&memento.used, memblk);
@@ -2594,6 +3010,7 @@ void *Memento_realloc(void *blk, size_t newsize)
         MEMENTO_LOCK();
         ret = do_malloc(newsize, Memento_EventType_realloc);
         MEMENTO_UNLOCK();
+        if (!ret) errno = ENOMEM;
         return ret;
     }
     if (newsize == 0) {
@@ -2606,6 +3023,7 @@ void *Memento_realloc(void *blk, size_t newsize)
     MEMENTO_LOCK();
     ret = do_realloc(blk, newsize, Memento_EventType_realloc);
     MEMENTO_UNLOCK();
+    if (!ret) errno = ENOMEM;
     return ret;
 }
 
@@ -3061,6 +3479,7 @@ void *(Memento_reference)(void *a)
 #undef Memento_free
 #undef Memento_realloc
 #undef Memento_calloc
+#undef Memento_strdup
 
 void *Memento_malloc(size_t size)
 {
@@ -3082,6 +3501,68 @@ void *Memento_calloc(size_t n, size_t s)
     return MEMENTO_UNDERLYING_CALLOC(n, s);
 }
 
+/* Avoid calling strdup, in case our compiler doesn't support it.
+ * Yes, I'm looking at you, early Visual Studios. */
+char *Memento_strdup(const char *s)
+{
+    size_t len = strlen(s)+1;
+    char *ret = MEMENTO_UNDERLYING_MALLOC(len);
+    if (ret != NULL)
+        memcpy(ret, s, len);
+    return ret;
+}
+
+/* Avoid calling asprintf, in case our compiler doesn't support it.
+ * Vaguely unhappy about relying on vsnprintf, but... */
+int Memento_asprintf(char **ret, const char *format, ...)
+{
+    va_list va;
+    int n;
+    int n2;
+
+    va_start(va, format);
+    n = vsnprintf(NULL, 0, format, va);
+    va_end(va);
+    if (n < 0)
+        return n;
+
+    *ret = MEMENTO_UNDERLYING_MALLOC(n+1);
+    if (*ret == NULL)
+        return -1;
+
+    va_start(va, format);
+    n2 = vsnprintf(*ret, n + 1, format, va);
+    va_end(va);
+
+    return n2;
+}
+
+/* Avoid calling vasprintf, in case our compiler doesn't support it.
+ * Vaguely unhappy about relying on vsnprintf, but... */
+int Memento_vasprintf(char **ret, const char *format, va_list ap)
+{
+    int n;
+    va_list ap2;
+    va_copy(ap2, ap);
+
+    n = vsnprintf(NULL, 0, format, ap);
+    if (n < 0) {
+        va_end(ap2);
+        return n;
+    }
+
+    *ret = MEMENTO_UNDERLYING_MALLOC(n+1);
+    if (*ret == NULL) {
+        va_end(ap2);
+        return -1;
+    }
+
+    n = vsnprintf(*ret, n + 1, format, ap2);
+    va_end(ap2);
+
+    return n;
+}
+
 void (Memento_listBlocks)(void)
 {
 }
@@ -3112,6 +3593,10 @@ void (Memento_listBlockInfo)(void)
 {
 }
 
+void (Memento_blockInfo)(void *ptr)
+{
+}
+
 void (Memento_startLeaking)(void)
 {
 }
