@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/ucdn.h"
 
@@ -234,7 +256,7 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_ARABIC:
 		if (language == FZ_LANG_ur || language == FZ_LANG_urd)
 			RETURN(noto,NotoNastaliqUrdu_Regular_otf);
-		RETURN(noto,NotoNaskhArabic_Regular_ttf);
+		RETURN(noto,NotoNaskhArabic_Regular_otf);
 
 	case UCDN_SCRIPT_ADLAM: RETURN(noto,NotoSansAdlam_Regular_otf);
 	case UCDN_SCRIPT_AHOM: RETURN(noto,NotoSerifAhom_Regular_otf);
@@ -245,7 +267,7 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_BAMUM: RETURN(noto,NotoSansBamum_Regular_otf);
 	case UCDN_SCRIPT_BASSA_VAH: RETURN(noto,NotoSansBassaVah_Regular_otf);
 	case UCDN_SCRIPT_BATAK: RETURN(noto,NotoSansBatak_Regular_otf);
-	case UCDN_SCRIPT_BENGALI: RETURN(noto,NotoSerifBengali_Regular_ttf);
+	case UCDN_SCRIPT_BENGALI: RETURN(noto,NotoSerifBengali_Regular_otf);
 	case UCDN_SCRIPT_BHAIKSUKI: RETURN(noto,NotoSansBhaiksuki_Regular_otf);
 	case UCDN_SCRIPT_BRAHMI: RETURN(noto,NotoSansBrahmi_Regular_otf);
 	case UCDN_SCRIPT_BUGINESE: RETURN(noto,NotoSansBuginese_Regular_otf);
@@ -256,25 +278,27 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_CHAKMA: RETURN(noto,NotoSansChakma_Regular_otf);
 	case UCDN_SCRIPT_CHAM: RETURN(noto,NotoSansCham_Regular_otf);
 	case UCDN_SCRIPT_CHEROKEE: RETURN(noto,NotoSansCherokee_Regular_otf);
+	case UCDN_SCRIPT_CHORASMIAN: break;
 	case UCDN_SCRIPT_COPTIC: RETURN(noto,NotoSansCoptic_Regular_otf);
 	case UCDN_SCRIPT_CUNEIFORM: RETURN(noto,NotoSansCuneiform_Regular_otf);
 	case UCDN_SCRIPT_CYPRIOT: RETURN(noto,NotoSansCypriot_Regular_otf);
 	case UCDN_SCRIPT_DESERET: RETURN(noto,NotoSansDeseret_Regular_otf);
-	case UCDN_SCRIPT_DEVANAGARI: RETURN(noto,NotoSerifDevanagari_Regular_ttf);
-	case UCDN_SCRIPT_DOGRA: break;
+	case UCDN_SCRIPT_DEVANAGARI: RETURN(noto,NotoSerifDevanagari_Regular_otf);
+	case UCDN_SCRIPT_DIVES_AKURU: break;
+	case UCDN_SCRIPT_DOGRA: RETURN(noto,NotoSerifDogra_Regular_otf);
 	case UCDN_SCRIPT_DUPLOYAN: RETURN(noto,NotoSansDuployan_Regular_otf);
 	case UCDN_SCRIPT_EGYPTIAN_HIEROGLYPHS: RETURN(noto,NotoSansEgyptianHieroglyphs_Regular_otf);
 	case UCDN_SCRIPT_ELBASAN: RETURN(noto,NotoSansElbasan_Regular_otf);
-	case UCDN_SCRIPT_ELYMAIC: break;
+	case UCDN_SCRIPT_ELYMAIC: RETURN(noto,NotoSansElymaic_Regular_otf);
 	case UCDN_SCRIPT_ETHIOPIC: RETURN(noto,NotoSerifEthiopic_Regular_otf);
 	case UCDN_SCRIPT_GEORGIAN: RETURN(noto,NotoSerifGeorgian_Regular_otf);
 	case UCDN_SCRIPT_GLAGOLITIC: RETURN(noto,NotoSansGlagolitic_Regular_otf);
 	case UCDN_SCRIPT_GOTHIC: RETURN(noto,NotoSansGothic_Regular_otf);
-	case UCDN_SCRIPT_GRANTHA: RETURN(noto,NotoSansGrantha_Regular_otf);
+	case UCDN_SCRIPT_GRANTHA: RETURN(noto,NotoSerifGrantha_Regular_otf);
 	case UCDN_SCRIPT_GUJARATI: RETURN(noto,NotoSerifGujarati_Regular_otf);
-	case UCDN_SCRIPT_GUNJALA_GONDI: break;
+	case UCDN_SCRIPT_GUNJALA_GONDI: RETURN(noto,NotoSansGunjalaGondi_Regular_otf);
 	case UCDN_SCRIPT_GURMUKHI: RETURN(noto,NotoSerifGurmukhi_Regular_otf);
-	case UCDN_SCRIPT_HANIFI_ROHINGYA: break;
+	case UCDN_SCRIPT_HANIFI_ROHINGYA: RETURN(noto,NotoSansHanifiRohingya_Regular_otf);
 	case UCDN_SCRIPT_HANUNOO: RETURN(noto,NotoSansHanunoo_Regular_otf);
 	case UCDN_SCRIPT_HATRAN: RETURN(noto,NotoSansHatran_Regular_otf);
 	case UCDN_SCRIPT_HEBREW: RETURN(noto,NotoSerifHebrew_Regular_otf);
@@ -287,7 +311,8 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_KAYAH_LI: RETURN(noto,NotoSansKayahLi_Regular_otf);
 	case UCDN_SCRIPT_KHAROSHTHI: RETURN(noto,NotoSansKharoshthi_Regular_otf);
 	case UCDN_SCRIPT_KHMER: RETURN(noto,NotoSerifKhmer_Regular_otf);
-	case UCDN_SCRIPT_KHOJKI: RETURN(noto,NotoSansKhojki_Regular_otf);;
+	case UCDN_SCRIPT_KHITAN_SMALL_SCRIPT: break;
+	case UCDN_SCRIPT_KHOJKI: RETURN(noto,NotoSerifKhojki_Regular_otf);;
 	case UCDN_SCRIPT_KHUDAWADI: RETURN(noto,NotoSansKhudawadi_Regular_otf);
 	case UCDN_SCRIPT_LAO: RETURN(noto,NotoSerifLao_Regular_otf);
 	case UCDN_SCRIPT_LEPCHA: RETURN(noto,NotoSansLepcha_Regular_otf);
@@ -299,12 +324,12 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_LYDIAN: RETURN(noto,NotoSansLydian_Regular_otf);
 	case UCDN_SCRIPT_MAHAJANI: RETURN(noto,NotoSansMahajani_Regular_otf);
 	case UCDN_SCRIPT_MAKASAR: break;
-	case UCDN_SCRIPT_MALAYALAM: RETURN(noto,NotoSerifMalayalam_Regular_ttf);
+	case UCDN_SCRIPT_MALAYALAM: RETURN(noto,NotoSerifMalayalam_Regular_otf);
 	case UCDN_SCRIPT_MANDAIC: RETURN(noto,NotoSansMandaic_Regular_otf);
 	case UCDN_SCRIPT_MANICHAEAN: RETURN(noto,NotoSansManichaean_Regular_otf);
 	case UCDN_SCRIPT_MARCHEN: RETURN(noto,NotoSansMarchen_Regular_otf);
-	case UCDN_SCRIPT_MASARAM_GONDI: break;
-	case UCDN_SCRIPT_MEDEFAIDRIN: break;
+	case UCDN_SCRIPT_MASARAM_GONDI: RETURN(noto,NotoSansMasaramGondi_Regular_otf);
+	case UCDN_SCRIPT_MEDEFAIDRIN: RETURN(noto,NotoSansMedefaidrin_Regular_otf);
 	case UCDN_SCRIPT_MEETEI_MAYEK: RETURN(noto,NotoSansMeeteiMayek_Regular_otf);
 	case UCDN_SCRIPT_MENDE_KIKAKUI: RETURN(noto,NotoSansMendeKikakui_Regular_otf);
 	case UCDN_SCRIPT_MEROITIC_CURSIVE: RETURN(noto,NotoSansMeroitic_Regular_otf);
@@ -320,19 +345,19 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_NEWA: RETURN(noto,NotoSansNewa_Regular_otf);
 	case UCDN_SCRIPT_NEW_TAI_LUE: RETURN(noto,NotoSansNewTaiLue_Regular_otf);
 	case UCDN_SCRIPT_NKO: RETURN(noto,NotoSansNKo_Regular_otf);
-	case UCDN_SCRIPT_NUSHU: break;
-	case UCDN_SCRIPT_NYIAKENG_PUACHUE_HMONG: break;
+	case UCDN_SCRIPT_NUSHU: RETURN(noto,NotoSansNushu_Regular_otf);
+	case UCDN_SCRIPT_NYIAKENG_PUACHUE_HMONG: RETURN(noto,NotoSerifNyiakengPuachueHmong_Regular_otf);
 	case UCDN_SCRIPT_OGHAM: RETURN(noto,NotoSansOgham_Regular_otf);
 	case UCDN_SCRIPT_OLD_HUNGARIAN: RETURN(noto,NotoSansOldHungarian_Regular_otf);
 	case UCDN_SCRIPT_OLD_ITALIC: RETURN(noto,NotoSansOldItalic_Regular_otf);
 	case UCDN_SCRIPT_OLD_NORTH_ARABIAN: RETURN(noto,NotoSansOldNorthArabian_Regular_otf);
 	case UCDN_SCRIPT_OLD_PERMIC: RETURN(noto,NotoSansOldPermic_Regular_otf);
 	case UCDN_SCRIPT_OLD_PERSIAN: RETURN(noto,NotoSansOldPersian_Regular_otf);
-	case UCDN_SCRIPT_OLD_SOGDIAN: break;
+	case UCDN_SCRIPT_OLD_SOGDIAN: RETURN(noto,NotoSansOldSogdian_Regular_otf);
 	case UCDN_SCRIPT_OLD_SOUTH_ARABIAN: RETURN(noto,NotoSansOldSouthArabian_Regular_otf);
 	case UCDN_SCRIPT_OLD_TURKIC: RETURN(noto,NotoSansOldTurkic_Regular_otf);
 	case UCDN_SCRIPT_OL_CHIKI: RETURN(noto,NotoSansOlChiki_Regular_otf);
-	case UCDN_SCRIPT_ORIYA: RETURN(noto,NotoSansOriya_Regular_ttf);
+	case UCDN_SCRIPT_ORIYA: RETURN(noto,NotoSansOriya_Regular_otf);
 	case UCDN_SCRIPT_OSAGE: RETURN(noto,NotoSansOsage_Regular_otf);
 	case UCDN_SCRIPT_OSMANYA: RETURN(noto,NotoSansOsmanya_Regular_otf);
 	case UCDN_SCRIPT_PAHAWH_HMONG: RETURN(noto,NotoSansPahawhHmong_Regular_otf);
@@ -350,20 +375,20 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_SIDDHAM: RETURN(noto,NotoSansSiddham_Regular_otf);
 	case UCDN_SCRIPT_SIGNWRITING: break;
 	case UCDN_SCRIPT_SINHALA: RETURN(noto,NotoSerifSinhala_Regular_otf);
-	case UCDN_SCRIPT_SOGDIAN: break;
+	case UCDN_SCRIPT_SOGDIAN: RETURN(noto,NotoSansSogdian_Regular_otf);
 	case UCDN_SCRIPT_SORA_SOMPENG: RETURN(noto,NotoSansSoraSompeng_Regular_otf);
-	case UCDN_SCRIPT_SOYOMBO: break;
+	case UCDN_SCRIPT_SOYOMBO: RETURN(noto,NotoSansSoyombo_Regular_otf);
 	case UCDN_SCRIPT_SUNDANESE: RETURN(noto,NotoSansSundanese_Regular_otf);
 	case UCDN_SCRIPT_SYLOTI_NAGRI: RETURN(noto,NotoSansSylotiNagri_Regular_otf);
 	case UCDN_SCRIPT_SYRIAC: RETURN(noto,NotoSansSyriac_Regular_otf);
 	case UCDN_SCRIPT_TAGALOG: RETURN(noto,NotoSansTagalog_Regular_otf);
 	case UCDN_SCRIPT_TAGBANWA: RETURN(noto,NotoSansTagbanwa_Regular_otf);
 	case UCDN_SCRIPT_TAI_LE: RETURN(noto,NotoSansTaiLe_Regular_otf);
-	case UCDN_SCRIPT_TAI_THAM: RETURN(noto,NotoSansTaiTham_Regular_ttf);
+	case UCDN_SCRIPT_TAI_THAM: RETURN(noto,NotoSansTaiTham_Regular_otf);
 	case UCDN_SCRIPT_TAI_VIET: RETURN(noto,NotoSansTaiViet_Regular_otf);
 	case UCDN_SCRIPT_TAKRI: RETURN(noto,NotoSansTakri_Regular_otf);
 	case UCDN_SCRIPT_TAMIL: RETURN(noto,NotoSerifTamil_Regular_otf);
-	case UCDN_SCRIPT_TELUGU: RETURN(noto,NotoSerifTelugu_Regular_ttf);
+	case UCDN_SCRIPT_TELUGU: RETURN(noto,NotoSerifTelugu_Regular_otf);
 	case UCDN_SCRIPT_THAANA: RETURN(noto,NotoSansThaana_Regular_otf);
 	case UCDN_SCRIPT_THAI: RETURN(noto,NotoSerifThai_Regular_otf);
 	case UCDN_SCRIPT_TIBETAN: RETURN(noto,NotoSerifTibetan_Regular_otf);
@@ -371,10 +396,11 @@ fz_lookup_noto_font(fz_context *ctx, int script, int language, int *size, int *s
 	case UCDN_SCRIPT_TIRHUTA: RETURN(noto,NotoSansTirhuta_Regular_otf);
 	case UCDN_SCRIPT_UGARITIC: RETURN(noto,NotoSansUgaritic_Regular_otf);
 	case UCDN_SCRIPT_VAI: RETURN(noto,NotoSansVai_Regular_otf);
-	case UCDN_SCRIPT_WANCHO: break;
+	case UCDN_SCRIPT_WANCHO: RETURN(noto,NotoSansWancho_Regular_otf);
 	case UCDN_SCRIPT_WARANG_CITI: RETURN(noto,NotoSansWarangCiti_Regular_otf);
+	case UCDN_SCRIPT_YEZIDI: RETURN(noto,NotoSerifYezidi_Regular_otf);
 	case UCDN_SCRIPT_YI: RETURN(noto,NotoSansYi_Regular_otf);
-	case UCDN_SCRIPT_ZANABAZAR_SQUARE: break;
+	case UCDN_SCRIPT_ZANABAZAR_SQUARE: RETURN(noto,NotoSansZanabazarSquare_Regular_otf);
 
 #if NOTO_TANGUT
 	case UCDN_SCRIPT_TANGUT: RETURN(noto,NotoSerifTangut_Regular_otf);
