@@ -1,6 +1,28 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
-#include <zlib.h>
+#include "z-imp.h"
 
 typedef struct ps_band_writer_s
 {
@@ -219,7 +241,7 @@ ps_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_sta
 	{
 		fz_free(ctx, writer->input);
 		writer->input = NULL;
-		writer->input = fz_malloc(ctx, required_input);
+		writer->input = Memento_label(fz_malloc(ctx, required_input), "pswriter_input");
 		writer->input_size = required_input;
 	}
 
@@ -227,7 +249,7 @@ ps_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_sta
 	{
 		fz_free(ctx, writer->output);
 		writer->output = NULL;
-		writer->output = fz_malloc(ctx, required_output);
+		writer->output = Memento_label(fz_malloc(ctx, required_output), "pswriter_output");
 		writer->output_size = required_output;
 	}
 
@@ -268,16 +290,14 @@ fz_band_writer *fz_new_ps_band_writer(fz_context *ctx, fz_output *out)
 
 /* High-level document writer interface */
 
-typedef struct fz_ps_writer_s fz_ps_writer;
-
-struct fz_ps_writer_s
+typedef struct
 {
 	fz_document_writer super;
 	fz_draw_options draw;
 	fz_pixmap *pixmap;
 	fz_output *out;
 	int count;
-};
+} fz_ps_writer;
 
 static fz_device *
 ps_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
@@ -329,22 +349,32 @@ ps_drop_writer(fz_context *ctx, fz_document_writer *wri_)
 }
 
 fz_document_writer *
-fz_new_ps_writer(fz_context *ctx, const char *path, const char *options)
+fz_new_ps_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
 {
-	fz_ps_writer *wri = fz_new_derived_document_writer(ctx, fz_ps_writer, ps_begin_page, ps_end_page, ps_close_writer, ps_drop_writer);
+	fz_ps_writer *wri = NULL;
+
+	fz_var(wri);
 
 	fz_try(ctx)
 	{
+		wri = fz_new_derived_document_writer(ctx, fz_ps_writer, ps_begin_page, ps_end_page, ps_close_writer, ps_drop_writer);
 		fz_parse_draw_options(ctx, &wri->draw, options);
-		wri->out = fz_new_output_with_path(ctx, path ? path : "out.ps", 0);
+		wri->out = out;
 		fz_write_ps_file_header(ctx, wri->out);
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_output(ctx, wri->out);
+		fz_drop_output(ctx, out);
 		fz_free(ctx, wri);
 		fz_rethrow(ctx);
 	}
 
 	return (fz_document_writer*)wri;
 }
+
+fz_document_writer *
+fz_new_ps_writer(fz_context *ctx, const char *path, const char *options)
+{
+	fz_output *out = fz_new_output_with_path(ctx, path ? path : "out.ps", 0);
+	return fz_new_ps_writer_with_output(ctx, out, options);
+}
