@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <assert.h>
@@ -8,9 +30,6 @@ typedef struct {
 	fz_pwg_options pwg;
 } pwg_band_writer;
 
-/*
-	Output the file header to a pwg stream, ready for pages to follow it.
-*/
 void
 fz_write_pwg_file_header(fz_context *ctx, fz_output *out)
 {
@@ -423,20 +442,123 @@ fz_band_writer *fz_new_pwg_band_writer(fz_context *ctx, fz_output *out, const fz
 
 const char *fz_pwg_write_options_usage =
 	"PWG output options:\n"
-	"\tcolorspace=mono: render 1-bit black and white bitmap\n"
+	"\tmedia_class=<string>: set the media_class field\n"
+	"\tmedia_color=<string>: set the media_color field\n"
+	"\tmedia_type=<string>: set the media_type field\n"
+	"\toutput_type=<string>: set the output_type field\n"
+	"\trendering_intent=<string>: set the rendering_intent field\n"
+	"\tpage_size_name=<string>: set the page_size_name field\n"
+	"\tadvance_distance=<int>: set the advance_distance field\n"
+	"\tadvance_media=<int>: set the advance_media field\n"
+	"\tcollate=<int>: set the collate field\n"
+	"\tcut_media=<int>: set the cut_media field\n"
+	"\tduplex=<int>: set the duplex field\n"
+	"\tinsert_sheet=<int>: set the insert_sheet field\n"
+	"\tjog=<int>: set the jog field\n"
+	"\tleading_edge=<int>: set the leading_edge field\n"
+	"\tmanual_feed=<int>: set the manual_feed field\n"
+	"\tmedia_position=<int>: set the media_position field\n"
+	"\tmedia_weight=<int>: set the media_weight field\n"
+	"\tmirror_print=<int>: set the mirror_print field\n"
+	"\tnegative_print=<int>: set the negative_print field\n"
+	"\tnum_copies=<int>: set the num_copies field\n"
+	"\torientation=<int>: set the orientation field\n"
+	"\toutput_face_up=<int>: set the output_face_up field\n"
+	"\tpage_size_x=<int>: set the page_size_x field\n"
+	"\tpage_size_y=<int>: set the page_size_y field\n"
+	"\tseparations=<int>: set the separations field\n"
+	"\ttray_switch=<int>: set the tray_switch field\n"
+	"\ttumble=<int>: set the tumble field\n"
+	"\tmedia_type_num=<int>: set the media_type_num field\n"
+	"\tcompression=<int>: set the compression field\n"
+	"\trow_count=<int>: set the row_count field\n"
+	"\trow_feed=<int>: set the row_feed field\n"
+	"\trow_step=<int>: set the row_step field\n"
 	"\n";
 
+static void
+warn_if_long(fz_context *ctx, const char *str, size_t ret)
+{
+	if (ret > 0)
+		fz_warn(ctx, "Option %s is too long, truncated.", str);
+}
+
 fz_pwg_options *
 fz_parse_pwg_options(fz_context *ctx, fz_pwg_options *opts, const char *args)
 {
+	const char *val;
+
 	memset(opts, 0, sizeof *opts);
 
+	if (fz_has_option(ctx, args, "media_class", &val))
+		warn_if_long(ctx, "media_class", fz_copy_option(ctx, val, opts->media_class, 64));
+	if (fz_has_option(ctx, args, "media_color", &val))
+		warn_if_long(ctx, "media_color", fz_copy_option(ctx, val, opts->media_color, 64));
+	if (fz_has_option(ctx, args, "media_type", &val))
+		warn_if_long(ctx, "media_type", fz_copy_option(ctx, val, opts->media_type, 64));
+	if (fz_has_option(ctx, args, "output_type", &val))
+		warn_if_long(ctx, "output_type", fz_copy_option(ctx, val, opts->output_type, 64));
+	if (fz_has_option(ctx, args, "rendering_intent", &val))
+		warn_if_long(ctx, "rendering_intent", fz_copy_option(ctx, val, opts->rendering_intent, 64));
+	if (fz_has_option(ctx, args, "page_size_name", &val))
+		warn_if_long(ctx, "page_size_name", fz_copy_option(ctx, val, opts->page_size_name, 64));
+	if (fz_has_option(ctx, args, "advance_distance", &val))
+		opts->advance_distance = fz_atoi(val);
+	if (fz_has_option(ctx, args, "advance_media", &val))
+		opts->advance_media = fz_atoi(val);
+	if (fz_has_option(ctx, args, "collate", &val))
+		opts->collate = fz_atoi(val);
+	if (fz_has_option(ctx, args, "cut_media", &val))
+		opts->cut_media = fz_atoi(val);
+	if (fz_has_option(ctx, args, "duplex", &val))
+		opts->duplex = fz_atoi(val);
+	if (fz_has_option(ctx, args, "insert_sheet", &val))
+		opts->insert_sheet = fz_atoi(val);
+	if (fz_has_option(ctx, args, "jog", &val))
+		opts->jog = fz_atoi(val);
+	if (fz_has_option(ctx, args, "leading_edge", &val))
+		opts->leading_edge = fz_atoi(val);
+	if (fz_has_option(ctx, args, "manual_feed", &val))
+		opts->manual_feed = fz_atoi(val);
+	if (fz_has_option(ctx, args, "media_position", &val))
+		opts->media_position = fz_atoi(val);
+	if (fz_has_option(ctx, args, "media_weight", &val))
+		opts->media_weight = fz_atoi(val);
+	if (fz_has_option(ctx, args, "mirror_print", &val))
+		opts->mirror_print = fz_atoi(val);
+	if (fz_has_option(ctx, args, "negative_print", &val))
+		opts->negative_print = fz_atoi(val);
+	if (fz_has_option(ctx, args, "num_copies", &val))
+		opts->num_copies = fz_atoi(val);
+	if (fz_has_option(ctx, args, "orientation", &val))
+		opts->orientation = fz_atoi(val);
+	if (fz_has_option(ctx, args, "output_face_up", &val))
+		opts->output_face_up = fz_atoi(val);
+	if (fz_has_option(ctx, args, "page_size_x", &val))
+		opts->PageSize[0] = fz_atoi(val);
+	if (fz_has_option(ctx, args, "page_size_y", &val))
+		opts->PageSize[1] = fz_atoi(val);
+	if (fz_has_option(ctx, args, "separations", &val))
+		opts->separations = fz_atoi(val);
+	if (fz_has_option(ctx, args, "tray_switch", &val))
+		opts->tray_switch = fz_atoi(val);
+	if (fz_has_option(ctx, args, "tumble", &val))
+		opts->tumble = fz_atoi(val);
+	if (fz_has_option(ctx, args, "media_type_num", &val))
+		opts->media_type_num = fz_atoi(val);
+	if (fz_has_option(ctx, args, "compression", &val))
+		opts->compression = fz_atoi(val);
+	if (fz_has_option(ctx, args, "row_count", &val))
+		opts->row_count = fz_atoi(val);
+	if (fz_has_option(ctx, args, "row_feed", &val))
+		opts->row_feed = fz_atoi(val);
+	if (fz_has_option(ctx, args, "row_step", &val))
+		opts->row_step = fz_atoi(val);
+
 	return opts;
 }
 
-typedef struct fz_pwg_writer_s fz_pwg_writer;
-
-struct fz_pwg_writer_s
+typedef struct
 {
 	fz_document_writer super;
 	fz_draw_options draw;
@@ -444,7 +566,7 @@ struct fz_pwg_writer_s
 	int mono;
 	fz_pixmap *pixmap;
 	fz_output *out;
-};
+} fz_pwg_writer;
 
 static fz_device *
 pwg_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
@@ -503,27 +625,37 @@ pwg_drop_writer(fz_context *ctx, fz_document_writer *wri_)
 }
 
 fz_document_writer *
-fz_new_pwg_writer(fz_context *ctx, const char *path, const char *options)
+fz_new_pwg_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
 {
-	fz_pwg_writer *wri = fz_new_derived_document_writer(ctx, fz_pwg_writer, pwg_begin_page, pwg_end_page, pwg_close_writer, pwg_drop_writer);
+	fz_pwg_writer *wri = NULL;
 	const char *val;
 
+	fz_var(wri);
+
 	fz_try(ctx)
 	{
+		wri = fz_new_derived_document_writer(ctx, fz_pwg_writer, pwg_begin_page, pwg_end_page, pwg_close_writer, pwg_drop_writer);
 		fz_parse_draw_options(ctx, &wri->draw, options);
 		fz_parse_pwg_options(ctx, &wri->pwg, options);
 		if (fz_has_option(ctx, options, "colorspace", &val))
 			if (fz_option_eq(val, "mono"))
 				wri->mono = 1;
-		wri->out = fz_new_output_with_path(ctx, path ? path : "out.pwg", 0);
+		wri->out = out;
 		fz_write_pwg_file_header(ctx, wri->out);
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_output(ctx, wri->out);
+		fz_drop_output(ctx, out);
 		fz_free(ctx, wri);
 		fz_rethrow(ctx);
 	}
 
 	return (fz_document_writer*)wri;
 }
+
+fz_document_writer *
+fz_new_pwg_writer(fz_context *ctx, const char *path, const char *options)
+{
+	fz_output *out = fz_new_output_with_path(ctx, path ? path : "out.pwg", 0);
+	return fz_new_pwg_writer_with_output(ctx, out, options);
+}
