@@ -1,24 +1,46 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <string.h>
 #include <stdio.h>
 
-typedef struct fz_pool_node_s fz_pool_node;
+typedef struct fz_pool_node
+{
+	struct fz_pool_node *next;
+	char mem[1];
+} fz_pool_node;
 
 #define POOL_SIZE (4<<10) /* default size of pool blocks */
 #define POOL_SELF (1<<10) /* size where allocs are put into their own blocks */
 
-struct fz_pool_s
+struct fz_pool
 {
+	size_t size;
 	fz_pool_node *head, *tail;
 	char *pos, *end;
 };
 
-struct fz_pool_node_s
-{
-	fz_pool_node *next;
-	char mem[1];
-};
 
 fz_pool *fz_new_pool(fz_context *ctx)
 {
@@ -28,7 +50,7 @@ fz_pool *fz_new_pool(fz_context *ctx)
 	pool = fz_malloc_struct(ctx, fz_pool);
 	fz_try(ctx)
 	{
-		node = fz_calloc(ctx, offsetof(fz_pool_node, mem) + POOL_SIZE, 1);
+		node = Memento_label(fz_calloc(ctx, offsetof(fz_pool_node, mem) + POOL_SIZE, 1), "fz_pool_block");
 		pool->head = pool->tail = node;
 		pool->pos = node->mem;
 		pool->end = node->mem + POOL_SIZE;
@@ -47,9 +69,10 @@ static void *fz_pool_alloc_oversize(fz_context *ctx, fz_pool *pool, size_t size)
 	fz_pool_node *node;
 
 	/* link in memory at the head of the list */
-	node = fz_calloc(ctx, offsetof(fz_pool_node, mem) + size, 1);
+	node = Memento_label(fz_calloc(ctx, offsetof(fz_pool_node, mem) + size, 1), "fz_pool_oversize");
 	node->next = pool->head;
 	pool->head = node;
+	pool->size += offsetof(fz_pool_node, mem) + size;
 
 	return node->mem;
 }
@@ -62,14 +85,15 @@ void *fz_pool_alloc(fz_context *ctx, fz_pool *pool, size_t size)
 		return fz_pool_alloc_oversize(ctx, pool, size);
 
 	/* round size to pointer alignment (we don't expect to use doubles) */
-	size = ((size + sizeof(void*) - 1) / sizeof(void*)) * sizeof(void*);
+	size = (size + FZ_POINTER_ALIGN_MOD - 1) & ~(FZ_POINTER_ALIGN_MOD-1);
 
 	if (pool->pos + size > pool->end)
 	{
-		fz_pool_node *node = fz_calloc(ctx, offsetof(fz_pool_node, mem) + POOL_SIZE, 1);
+		fz_pool_node *node = Memento_label(fz_calloc(ctx, offsetof(fz_pool_node, mem) + POOL_SIZE, 1), "fz_pool_block");
 		pool->tail = pool->tail->next = node;
 		pool->pos = node->mem;
 		pool->end = node->mem + POOL_SIZE;
+		pool->size += offsetof(fz_pool_node, mem) + POOL_SIZE;
 	}
 	ptr = pool->pos;
 	pool->pos += size;
@@ -84,6 +108,11 @@ char *fz_pool_strdup(fz_context *ctx, fz_pool *pool, const char *s)
 	return p;
 }
 
+size_t fz_pool_size(fz_context *ctx, fz_pool *pool)
+{
+	return pool ? pool->size : 0;
+}
+
 void fz_drop_pool(fz_context *ctx, fz_pool *pool)
 {
 	fz_pool_node *node;
