@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <string.h>
@@ -137,7 +159,7 @@ fz_dirname(char *dir, const char *path, size_t n)
 
 #ifdef _WIN32
 
-char *fz_realpath(const char *path, char buf[PATH_MAX])
+char *fz_realpath(const char *path, char *buf)
 {
 	wchar_t wpath[PATH_MAX];
 	wchar_t wbuf[PATH_MAX];
@@ -156,7 +178,7 @@ char *fz_realpath(const char *path, char buf[PATH_MAX])
 
 #else
 
-char *fz_realpath(const char *path, char buf[PATH_MAX])
+char *fz_realpath(const char *path, char *buf)
 {
 	return realpath(path, buf);
 }
@@ -474,6 +496,37 @@ fz_runelen(int c)
 	return fz_runetochar(str, c);
 }
 
+int
+fz_runeidx(const char *s, const char *p)
+{
+	int rune;
+	int i = 0;
+	while (s < p) {
+		if (*(unsigned char *)s < Runeself)
+			++s;
+		else
+			s += fz_chartorune(&rune, s);
+		++i;
+	}
+	return i;
+}
+
+const char *
+fz_runeptr(const char *s, int i)
+{
+	int rune;
+	while (i-- > 0) {
+		rune = *(unsigned char*)s;
+		if (rune < Runeself) {
+			if (rune == 0)
+				return NULL;
+			++s;
+		} else
+			s += fz_chartorune(&rune, s);
+	}
+	return s;
+}
+
 int
 fz_utflen(const char *s)
 {
