@@ -1,46 +1,61 @@
-#include "fitz-imp.h"
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
+#include "mupdf/fitz.h"
 
 #include <string.h>
 #include <float.h>
 #include <math.h>
 
-typedef struct svg_device_s svg_device;
-
-typedef struct tile_s tile;
-typedef struct font_s font;
-typedef struct glyph_s glyph;
-typedef struct image_s image;
-
-struct tile_s
+typedef struct
 {
 	int pattern;
 	fz_matrix ctm;
 	fz_rect view;
 	fz_rect area;
 	fz_point step;
-};
+} tile;
 
-struct glyph_s
+typedef struct glyph
 {
 	float x_off;
 	float y_off;
-};
+} glyph;
 
-struct font_s
+typedef struct
 {
 	int id;
 	fz_font *font;
 	int max_sentlist;
 	glyph *sentlist;
-};
+} font;
 
-struct image_s
+typedef struct
 {
 	int id;
 	fz_image *image;
-};
+} image;
 
-struct svg_device_s
+typedef struct
 {
 	fz_device super;
 
@@ -71,7 +86,7 @@ struct svg_device_s
 	image *images;
 
 	int layers;
-};
+} svg_device;
 
 /* SVG is awkward about letting us define things within symbol definitions
  * so we have to delay definitions until after the symbol definition ends. */
@@ -472,7 +487,7 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 		if (fnt->sentlist[gid].x_off == FLT_MIN)
 		{
 			/* Need to send this one */
-			fz_rect rect;
+			fz_rect rect = fz_empty_rect;
 			fz_path *path;
 			out = start_def(ctx, sdev);
 			fz_write_printf(ctx, out, "<symbol id=\"font_%x_%x\">\n", fnt->id, gid);
@@ -501,6 +516,7 @@ svg_dev_text_span_as_paths_defs(fz_context *ctx, fz_device *dev, fz_text_span *s
 				shift.e = -rect.x0;
 				shift.f = -rect.y0;
 				fz_run_t3_glyph(ctx, span->font, gid, shift, dev);
+				fnt = &sdev->fonts[font_idx]; /* recursion may realloc the font array! */
 			}
 			fz_write_printf(ctx, out, "</symbol>\n");
 			out = end_def(ctx, sdev);
@@ -922,7 +938,7 @@ svg_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix c
 
 	fz_try(ctx)
 	{
-		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL);
+		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL, NULL);
 		if (alpha != 1.0f)
 			fz_write_printf(ctx, out, "<g opacity=\"%g\">\n", alpha);
 		fz_write_printf(ctx, out, "<image x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" xlink:href=\"", pix->x, pix->y, pix->w, pix->h);
@@ -1134,7 +1150,7 @@ svg_dev_begin_tile(fz_context *ctx, fz_device *dev, fz_rect area, fz_rect view,
 	/* The first thing we do is to capture the contents of the pattern
 	 * as a symbol we can reuse. */
 	out = start_def(ctx, sdev);
-	fz_write_printf(ctx, out, "<symbol id=\"pac%d\">\n", t->pattern);
+	fz_write_printf(ctx, out, "<symbol id=\"pac%d\" style=\"overflow:visible\">\n", t->pattern);
 
 	return 0;
 }
@@ -1273,22 +1289,6 @@ svg_dev_drop_device(fz_context *ctx, fz_device *dev)
 	fz_free(ctx, sdev->images);
 }
 
-/*
-	Create a device that outputs (single page)
-		SVG files to the given output stream.
-
-	output: The output stream to send the constructed SVG page to.
-
-	page_width, page_height: The page dimensions to use (in points).
-
-	text_format: How to emit text. One of the following values:
-		FZ_SVG_TEXT_AS_TEXT: As <text> elements with possible layout errors and mismatching fonts.
-		FZ_SVG_TEXT_AS_PATH: As <path> elements with exact visual appearance.
-
-	reuse_images: Share image resources using <symbol> definitions.
-
-	id: ID parameter to keep generated IDs unique across SVG files.
-*/
 fz_device *fz_new_svg_device_with_id(fz_context *ctx, fz_output *out, float page_width, float page_height, int text_format, int reuse_images, int *id)
 {
 	svg_device *dev = fz_new_derived_device(ctx, svg_device);
