@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include "color-imp.h"
@@ -231,14 +253,15 @@ static void fz_test_fill_compressed_8bpc_image(fz_context *ctx, fz_test_device *
 static void
 fz_test_fill_other_image(fz_context *ctx, fz_test_device *dev, fz_pixmap *pix, fz_color_params color_params)
 {
-	unsigned int count, i, k, h, sa, ss;
+	unsigned int count, i, k, h, sa;
+	size_t ss;
 	unsigned char *s;
 
 	count = pix->w;
 	h = pix->h;
 	s = pix->samples;
 	sa = pix->alpha;
-	ss = pix->stride - pix->w * pix->n;
+	ss = pix->stride - pix->w * (size_t)pix->n;
 
 	if (pix->colorspace == fz_device_rgb(ctx))
 	{
