@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #ifndef SOURCE_HTML_IMP_H
 #define SOURCE_HTML_IMP_H
 
@@ -10,7 +32,6 @@ typedef struct fz_css_style_splay_s fz_css_style_splay;
 
 typedef struct fz_css_s fz_css;
 typedef struct fz_css_rule_s fz_css_rule;
-typedef struct fz_css_match_prop_s fz_css_match_prop;
 typedef struct fz_css_match_s fz_css_match;
 typedef struct fz_css_style_s fz_css_style;
 
@@ -83,7 +104,7 @@ struct fz_css_condition_s
 
 struct fz_css_property_s
 {
-	char *name;
+	int name;
 	fz_css_value *value;
 	short spec;
 	short important;
@@ -98,18 +119,80 @@ struct fz_css_value_s
 	fz_css_value *next;
 };
 
-struct fz_css_match_prop_s
+enum
 {
-	const char *name; /* not owned */
-	fz_css_value *value; /* not owned */
-	int spec;
+	PRO_BACKGROUND_COLOR,
+	PRO_BORDER_BOTTOM_COLOR,
+	PRO_BORDER_BOTTOM_STYLE,
+	PRO_BORDER_BOTTOM_WIDTH,
+	PRO_BORDER_LEFT_COLOR,
+	PRO_BORDER_LEFT_STYLE,
+	PRO_BORDER_LEFT_WIDTH,
+	PRO_BORDER_RIGHT_COLOR,
+	PRO_BORDER_RIGHT_STYLE,
+	PRO_BORDER_RIGHT_WIDTH,
+	PRO_BORDER_TOP_COLOR,
+	PRO_BORDER_TOP_STYLE,
+	PRO_BORDER_TOP_WIDTH,
+	PRO_COLOR,
+	PRO_DIRECTION,
+	PRO_DISPLAY,
+	PRO_FONT_FAMILY,
+	PRO_FONT_SIZE,
+	PRO_FONT_STYLE,
+	PRO_FONT_VARIANT,
+	PRO_FONT_WEIGHT,
+	PRO_HEIGHT,
+	PRO_LETTER_SPACING,
+	PRO_LINE_HEIGHT,
+	PRO_LIST_STYLE_IMAGE,
+	PRO_LIST_STYLE_POSITION,
+	PRO_LIST_STYLE_TYPE,
+	PRO_MARGIN_BOTTOM,
+	PRO_MARGIN_LEFT,
+	PRO_MARGIN_RIGHT,
+	PRO_MARGIN_TOP,
+	PRO_ORPHANS,
+	PRO_PADDING_BOTTOM,
+	PRO_PADDING_LEFT,
+	PRO_PADDING_RIGHT,
+	PRO_PADDING_TOP,
+	PRO_PAGE_BREAK_AFTER,
+	PRO_PAGE_BREAK_BEFORE,
+	PRO_QUOTES,
+	PRO_SRC,
+	PRO_TEXT_ALIGN,
+	PRO_TEXT_INDENT,
+	PRO_TEXT_TRANSFORM,
+	PRO_VERTICAL_ALIGN,
+	PRO_VISIBILITY,
+	PRO_WHITE_SPACE,
+	PRO_WIDOWS,
+	PRO_WIDTH,
+	PRO_WORD_SPACING,
+
+	/* Number of real properties. */
+	NUM_PROPERTIES,
+
+	/* Short-hand properties (always expanded when applied, never used as is): */
+	PRO_BORDER,
+	PRO_BORDER_BOTTOM,
+	PRO_BORDER_COLOR,
+	PRO_BORDER_LEFT,
+	PRO_BORDER_RIGHT,
+	PRO_BORDER_STYLE,
+	PRO_BORDER_TOP,
+	PRO_BORDER_WIDTH,
+	PRO_LIST_STYLE,
+	PRO_MARGIN,
+	PRO_PADDING,
 };
 
 struct fz_css_match_s
 {
 	fz_css_match *up;
-	int count;
-	fz_css_match_prop prop[64];
+	short spec[NUM_PROPERTIES];
+	fz_css_value *value[NUM_PROPERTIES];
 };
 
 enum { DIS_NONE, DIS_BLOCK, DIS_INLINE, DIS_LIST_ITEM, DIS_INLINE_BLOCK, DIS_TABLE, DIS_TABLE_ROW, DIS_TABLE_CELL };
@@ -195,7 +278,6 @@ struct fz_css_style_splay_s {
 enum
 {
 	BOX_BLOCK,	/* block-level: contains block, break, flow, and table boxes */
-	BOX_BREAK,	/* block-level: empty <br> tag boxes */
 	BOX_FLOW,	/* block-level: contains only inline boxes */
 	BOX_INLINE,	/* inline-level: contains only inline boxes */
 	BOX_TABLE,	/* table: contains table-row */
@@ -289,8 +371,9 @@ void fz_parse_css(fz_context *ctx, fz_css *css, const char *source, const char *
 fz_css_property *fz_parse_css_properties(fz_context *ctx, fz_pool *pool, const char *source);
 void fz_drop_css(fz_context *ctx, fz_css *css);
 void fz_debug_css(fz_context *ctx, fz_css *css);
+const char *fz_css_property_name(int name);
 
-void fz_match_css(fz_context *ctx, fz_css_match *match, fz_css *css, fz_xml *node);
+void fz_match_css(fz_context *ctx, fz_css_match *match, fz_css_match *up, fz_css *css, fz_xml *node);
 void fz_match_css_at_page(fz_context *ctx, fz_css_match *match, fz_css *css);
 
 int fz_get_css_match_display(fz_css_match *node);
@@ -315,13 +398,16 @@ void fz_drop_html_font_set(fz_context *ctx, fz_html_font_set *htx);
 
 void fz_add_css_font_faces(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, const char *base_uri, fz_css *css);
 
-fz_html *fz_parse_html(fz_context *ctx, fz_html_font_set *htx, fz_archive *zip, const char *base_uri, fz_buffer *buf, const char *user_css);
+fz_html *fz_parse_fb2(fz_context *ctx, fz_html_font_set *htx, fz_archive *zip, const char *base_uri, fz_buffer *buf, const char *user_css);
+fz_html *fz_parse_html5(fz_context *ctx, fz_html_font_set *htx, fz_archive *zip, const char *base_uri, fz_buffer *buf, const char *user_css);
+fz_html *fz_parse_xhtml(fz_context *ctx, fz_html_font_set *htx, fz_archive *zip, const char *base_uri, fz_buffer *buf, const char *user_css);
+
 void fz_layout_html(fz_context *ctx, fz_html *html, float w, float h, float em);
 void fz_draw_html(fz_context *ctx, fz_device *dev, fz_matrix ctm, fz_html *html, int page);
 fz_outline *fz_load_html_outline(fz_context *ctx, fz_html *node);
 
 float fz_find_html_target(fz_context *ctx, fz_html *html, const char *id);
-fz_link *fz_load_html_links(fz_context *ctx, fz_html *html, int page, const char *base_uri, void *doc);
+fz_link *fz_load_html_links(fz_context *ctx, fz_html *html, int page, const char *base_uri);
 fz_html *fz_keep_html(fz_context *ctx, fz_html *html);
 void fz_drop_html(fz_context *ctx, fz_html *html);
 fz_bookmark fz_make_html_bookmark(fz_context *ctx, fz_html *html, int page);
