@@ -1,8 +1,28 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
-#include "../fitz/fitz-imp.h"
-
 #include <assert.h>
 
 #include <ft2build.h>
@@ -138,7 +158,7 @@ static int strcmp_ignore_space(const char *a, const char *b)
 const char *pdf_clean_font_name(const char *fontname)
 {
 	int i, k;
-	for (i = 0; i < nelem(base_font_names); i++)
+	for (i = 0; i < (int)nelem(base_font_names); i++)
 		for (k = 0; base_font_names[i][k]; k++)
 			if (!strcmp_ignore_space(base_font_names[i][k], fontname))
 				return base_font_names[i][0];
@@ -287,7 +307,7 @@ static const struct { int code; const char *name; } mre_diff_table[] =
 static int lookup_mre_code(const char *name)
 {
 	int i;
-	for (i = 0; i < nelem(mre_diff_table); ++i)
+	for (i = 0; i < (int)nelem(mre_diff_table); ++i)
 		if (!strcmp(name, mre_diff_table[i].name))
 			return mre_diff_table[i].code;
 	for (i = 0; i < 256; i++)
@@ -404,6 +424,8 @@ pdf_load_substitute_cjk_font(fz_context *ctx, pdf_font_desc *fontdesc, const cha
 
 	fontdesc->font->flags.ft_substitute = 1;
 	fontdesc->font->flags.ft_stretch = 0;
+	fontdesc->font->flags.cjk = 1;
+	fontdesc->font->flags.cjk_lang = ros;
 }
 
 static void
@@ -609,6 +631,19 @@ select_unknown_cmap(FT_Face face)
 	return NULL;
 }
 
+static int use_s22pdf_workaround(fz_context *ctx, pdf_obj *dict, pdf_obj *descriptor)
+{
+	if (descriptor)
+	{
+		if (pdf_dict_get(ctx, dict, PDF_NAME(Encoding)) != PDF_NAME(WinAnsiEncoding))
+			return 0;
+		if (pdf_dict_get_int(ctx, descriptor, PDF_NAME(Flags)) != 4)
+			return 0;
+		return 1;
+	}
+	return 0;
+}
+
 static pdf_font_desc *
 pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 {
@@ -649,10 +684,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 			pdf_load_builtin_font(ctx, fontdesc, basefont, 0);
 
 		/* Some chinese documents mistakenly consider WinAnsiEncoding to be codepage 936 */
-		if (descriptor && pdf_is_string(ctx, pdf_dict_get(ctx, descriptor, PDF_NAME(FontName))) &&
-			!pdf_dict_get(ctx, dict, PDF_NAME(ToUnicode)) &&
-			pdf_name_eq(ctx, pdf_dict_get(ctx, dict, PDF_NAME(Encoding)), PDF_NAME(WinAnsiEncoding)) &&
-			pdf_dict_get_int(ctx, descriptor, PDF_NAME(Flags)) == 4)
+		if (use_s22pdf_workaround(ctx, dict, descriptor))
 		{
 			char *cp936fonts[] = {
 				"\xCB\xCE\xCC\xE5", "SimSun,Regular",
@@ -686,6 +718,9 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		/* Encoding */
 
 		symbolic = fontdesc->flags & 4;
+		/* Bug 703273: If non-symbolic, we're not symbolic. */
+		if (fontdesc->flags & 32)
+			symbolic = 0;
 
 		if (kind == TYPE1)
 			cmap = select_type1_cmap(face);
@@ -703,7 +738,8 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		else
 			fz_warn(ctx, "freetype could not find any cmaps");
 
-		etable = fz_malloc_array(ctx, 256, unsigned short);
+		/* FIXME: etable may leak on error. */
+		etable = Memento_label(fz_malloc_array(ctx, 256, unsigned short), "cid_to_gid");
 		fontdesc->size += 256 * sizeof(unsigned short);
 		for (i = 0; i < 256; i++)
 		{
@@ -737,7 +773,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 						item = pdf_array_get(ctx, diff, i);
 						if (pdf_is_int(ctx, item))
 							k = pdf_to_int(ctx, item);
-						if (pdf_is_name(ctx, item) && k >= 0 && k < nelem(estrings))
+						if (pdf_is_name(ctx, item) && k >= 0 && k < (int)nelem(estrings))
 							estrings[k++] = pdf_to_name(ctx, item);
 					}
 				}
@@ -946,7 +982,7 @@ hail_mary_cmp_key(fz_context *ctx, void *k0, void *k1)
 }
 
 static void
-hail_mary_format_key(fz_context *ctx, char *s, int n, void *key_)
+hail_mary_format_key(fz_context *ctx, char *s, size_t n, void *key_)
 {
 	fz_strlcpy(s, "(hail mary font)", n);
 }
@@ -955,6 +991,7 @@ static int hail_mary_store_key; /* Dummy */
 
 static const fz_store_type hail_mary_store_type =
 {
+	"hail-mary",
 	hail_mary_make_hash_key,
 	hail_mary_keep_key,
 	hail_mary_drop_key,
@@ -1018,12 +1055,17 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 			const char *reg, *ord;
 
 			cidinfo = pdf_dict_get(ctx, dict, PDF_NAME(CIDSystemInfo));
-			if (!cidinfo)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cid font is missing info");
-
-			reg = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Registry), NULL);
-			ord = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Ordering), NULL);
-			fz_snprintf(collection, sizeof collection, "%s-%s", reg, ord);
+			if (cidinfo)
+			{
+				reg = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Registry), NULL);
+				ord = pdf_dict_get_string(ctx, cidinfo, PDF_NAME(Ordering), NULL);
+				fz_snprintf(collection, sizeof collection, "%s-%s", reg, ord);
+			}
+			else
+			{
+				fz_warn(ctx, "CIDFont is missing CIDSystemInfo dictionary; assuming Adobe-Identity");
+				fz_strlcpy(collection, "Adobe-Identity", sizeof collection);
+			}
 		}
 
 		/* Encoding */
@@ -1069,7 +1111,7 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 
 			len = fz_buffer_storage(ctx, buf, &data);
 			fontdesc->cid_to_gid_len = len / 2;
-			fontdesc->cid_to_gid = fz_malloc_array(ctx, fontdesc->cid_to_gid_len, unsigned short);
+			fontdesc->cid_to_gid = Memento_label(fz_malloc_array(ctx, fontdesc->cid_to_gid_len, unsigned short), "cid_to_gid_map");
 			fontdesc->size += fontdesc->cid_to_gid_len * sizeof(unsigned short);
 			for (z = 0; z < fontdesc->cid_to_gid_len; z++)
 				fontdesc->cid_to_gid[z] = (data[z * 2] << 8) + data[z * 2 + 1];
@@ -1330,7 +1372,7 @@ pdf_make_width_table(fz_context *ctx, pdf_font_desc *fontdesc)
 	}
 
 	font->width_count = n + 1;
-	font->width_table = fz_malloc_array(ctx, font->width_count, short);
+	font->width_table = Memento_label(fz_malloc_array(ctx, font->width_count, short), "font_widths");
 	fontdesc->size += font->width_count * sizeof(short);
 
 	font->width_default = fontdesc->dhmtx.w;
@@ -1404,7 +1446,7 @@ pdf_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *dict)
 		fontdesc = pdf_load_simple_font(ctx, doc, dict);
 	}
 
-	pdf_mark_obj(ctx, dict);
+	(void)pdf_mark_obj(ctx, dict);
 	fz_try(ctx)
 	{
 		/* Create glyph width table for stretching substitute fonts and text extraction. */
