@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -10,8 +32,7 @@
  * Emit graphics calls to device.
  */
 
-typedef struct pdf_material_s pdf_material;
-typedef struct pdf_run_processor_s pdf_run_processor;
+typedef struct pdf_run_processor pdf_run_processor;
 
 static void pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj *page_resources, fz_matrix transform, int is_smask);
 
@@ -29,7 +50,7 @@ enum
 	PDF_MAT_SHADE,
 };
 
-struct pdf_material_s
+typedef struct
 {
 	int kind;
 	fz_colorspace *colorspace;
@@ -39,9 +60,9 @@ struct pdf_material_s
 	fz_color_params color_params;
 	float alpha;
 	float v[FZ_MAX_COLORS];
-};
+} pdf_material;
 
-struct pdf_gstate_s
+struct pdf_gstate
 {
 	fz_matrix ctm;
 	int clip_depth;
@@ -65,7 +86,7 @@ struct pdf_gstate_s
 	int luminosity;
 };
 
-struct pdf_run_processor_s
+struct pdf_run_processor
 {
 	pdf_processor super;
 	fz_device *dev;
@@ -89,14 +110,12 @@ struct pdf_run_processor_s
 	int gparent;
 };
 
-typedef struct softmask_save_s softmask_save;
-
-struct softmask_save_s
+typedef struct
 {
 	pdf_obj *softmask;
 	pdf_obj *page_resources;
 	fz_matrix ctm;
-};
+} softmask_save;
 
 static pdf_gstate *
 begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
@@ -211,13 +230,22 @@ pdf_show_shade(fz_context *ctx, pdf_run_processor *pr, fz_shade *shd)
 
 	bbox = fz_bound_shade(ctx, shd, gstate->ctm);
 
-	gstate = pdf_begin_group(ctx, pr, bbox, &softmask);
+	fz_try(ctx)
+	{
+		gstate = pdf_begin_group(ctx, pr, bbox, &softmask);
 
-	/* FIXME: The gstate->ctm in the next line may be wrong; maybe
-	 * it should be the parent gstates ctm? */
-	fz_fill_shade(ctx, pr->dev, shd, gstate->ctm, gstate->fill.alpha, gstate->fill.color_params);
+		/* FIXME: The gstate->ctm in the next line may be wrong; maybe
+		 * it should be the parent gstates ctm? */
+		fz_fill_shade(ctx, pr->dev, shd, gstate->ctm, gstate->fill.alpha, gstate->fill.color_params);
 
-	pdf_end_group(ctx, pr, &softmask);
+		pdf_end_group(ctx, pr, &softmask);
+	}
+	fz_catch(ctx)
+	{
+		pdf_drop_obj(ctx, softmask.softmask);
+		pdf_drop_obj(ctx, softmask.page_resources);
+		fz_rethrow(ctx);
+	}
 }
 
 static pdf_material *
@@ -464,6 +492,11 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 		{
 			for (x = x0; x < x1; x++)
 			{
+				/* Calls to pdf_process_contents may cause the
+				 * gstate array to be realloced to be larger.
+				 * That can invalidate gstate. Hence reload
+				 * it each time round the loop. */
+				gstate = pr->gstate + pr->gtop;
 				gstate->ctm = fz_pre_translate(ptm, x * pat->xstep, y * pat->ystep);
 				pdf_gsave(ctx, pr);
 				pdf_process_contents(ctx, (pdf_processor*)pr, pat->document, pat->resources, pat->contents, NULL);
@@ -540,9 +573,18 @@ pdf_show_image(fz_context *ctx, pdf_run_processor *pr, fz_image *image)
 	else
 	{
 		softmask_save softmask = { NULL };
-		gstate = pdf_begin_group(ctx, pr, bbox, &softmask);
-		pdf_show_image_imp(ctx, pr, image, image_ctm, bbox);
-		pdf_end_group(ctx, pr, &softmask);
+		fz_try(ctx)
+		{
+			gstate = pdf_begin_group(ctx, pr, bbox, &softmask);
+			pdf_show_image_imp(ctx, pr, image, image_ctm, bbox);
+			pdf_end_group(ctx, pr, &softmask);
+		}
+		fz_catch(ctx)
+		{
+			pdf_drop_obj(ctx, softmask.softmask);
+			pdf_drop_obj(ctx, softmask.page_resources);
+			fz_rethrow(ctx);
+		}
 	}
 }
 
@@ -680,6 +722,8 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 	}
 	fz_catch(ctx)
 	{
+		pdf_drop_obj(ctx, softmask.softmask);
+		pdf_drop_obj(ctx, softmask.page_resources);
 		fz_rethrow(ctx);
 	}
 }
@@ -831,6 +875,8 @@ pdf_flush_text(fz_context *ctx, pdf_run_processor *pr)
 	}
 	fz_catch(ctx)
 	{
+		pdf_drop_obj(ctx, softmask.softmask);
+		pdf_drop_obj(ctx, softmask.page_resources);
 		fz_rethrow(ctx);
 	}
 
@@ -909,7 +955,7 @@ pdf_show_space(fz_context *ctx, pdf_run_processor *pr, float tadj)
 }
 
 static void
-show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, int len)
+show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, size_t len)
 {
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
 	pdf_font_desc *fontdesc = gstate->text.font;
@@ -928,12 +974,16 @@ show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, int len)
 		else
 			fz_warn(ctx, "cannot encode character");
 		if (cpt == 32 && w == 1)
+		{
+			/* Bug 703151: pdf_show_char can realloc gstate. */
+			gstate = pr->gstate + pr->gtop;
 			pdf_show_space(ctx, pr, gstate->text.word_space);
+		}
 	}
 }
 
 static void
-pdf_show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, int len)
+pdf_show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, size_t len)
 {
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
 	pdf_font_desc *fontdesc = gstate->text.font;
@@ -1467,6 +1517,7 @@ static void pdf_run_q(fz_context *ctx, pdf_processor *proc)
 static void pdf_run_Q(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	pdf_flush_text(ctx, pr);
 	pdf_grestore(ctx, pr);
 }
 
@@ -1713,13 +1764,13 @@ static void pdf_run_TJ(fz_context *ctx, pdf_processor *proc, pdf_obj *obj)
 	pdf_show_text(ctx, pr, obj);
 }
 
-static void pdf_run_Tj(fz_context *ctx, pdf_processor *proc, char *string, int string_len)
+static void pdf_run_Tj(fz_context *ctx, pdf_processor *proc, char *string, size_t string_len)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pdf_show_string(ctx, pr, (unsigned char *)string, string_len);
 }
 
-static void pdf_run_squote(fz_context *ctx, pdf_processor *proc, char *string, int string_len)
+static void pdf_run_squote(fz_context *ctx, pdf_processor *proc, char *string, size_t string_len)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
@@ -1727,7 +1778,7 @@ static void pdf_run_squote(fz_context *ctx, pdf_processor *proc, char *string, i
 	pdf_show_string(ctx, pr, (unsigned char*)string, string_len);
 }
 
-static void pdf_run_dquote(fz_context *ctx, pdf_processor *proc, float aw, float ac, char *string, int string_len)
+static void pdf_run_dquote(fz_context *ctx, pdf_processor *proc, float aw, float ac, char *string, size_t string_len)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
@@ -2168,7 +2219,7 @@ pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char
 		proc->path = fz_new_path(ctx);
 
 		proc->gcap = 64;
-		proc->gstate = fz_calloc(ctx, proc->gcap, sizeof(pdf_gstate));
+		proc->gstate = fz_malloc_struct_array(ctx, proc->gcap, pdf_gstate);
 
 		proc->gtop = 0;
 		pdf_init_gstate(ctx, &proc->gstate[0], ctm);
