@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "svg-imp.h"
 
@@ -185,9 +207,12 @@ svg_parse_transform(fz_context *ctx, svg_document *doc, const char *str, fz_matr
 
 		else if (!strcmp(keyword, "translate"))
 		{
-			if (nargs != 2)
+			if (nargs == 1)
+				transform = fz_concat(fz_translate(args[0], 0), transform);
+			else if (nargs == 2)
+				transform = fz_concat(fz_translate(args[0], args[1]), transform);
+			else
 				fz_throw(ctx, FZ_ERROR_SYNTAX, "wrong number of arguments to translate(): %d", nargs);
-			transform = fz_concat(fz_translate(args[0], args[1]), transform);
 		}
 
 		else if (!strcmp(keyword, "scale"))
@@ -202,9 +227,16 @@ svg_parse_transform(fz_context *ctx, svg_document *doc, const char *str, fz_matr
 
 		else if (!strcmp(keyword, "rotate"))
 		{
-			if (nargs != 1)
+			if (nargs == 1)
+				transform = fz_concat(fz_rotate(args[0]), transform);
+			else if (nargs == 3)
+			{
+				transform = fz_concat(fz_translate(args[1], args[2]), transform);
+				transform = fz_concat(fz_rotate(args[0]), transform);
+				transform = fz_concat(fz_translate(-args[1], -args[2]), transform);
+			}
+			else
 				fz_throw(ctx, FZ_ERROR_SYNTAX, "wrong number of arguments to rotate(): %d", nargs);
-			transform = fz_concat(fz_rotate(args[0]), transform);
 		}
 
 		else if (!strcmp(keyword, "skewX"))
@@ -229,3 +261,95 @@ svg_parse_transform(fz_context *ctx, svg_document *doc, const char *str, fz_matr
 
 	return transform;
 }
+
+float
+svg_parse_number_from_style(fz_context *ctx, svg_document *doc, const char *style, const char *att, float number)
+{
+	if (style)
+	{
+		char *end, *p = strstr(style, att);
+		if (p)
+		{
+			size_t n = strlen(att);
+			if (p[n] == ':')
+			{
+				p += n + 1;
+				while (*p && svg_is_whitespace(*p))
+					++p;
+				number = fz_strtof(p, &end);
+				if (end[0] == 'i' && end[1] == 'n') return number * 72;
+				if (end[0] == 'c' && end[1] == 'm') return number * 7200 / 254;
+				if (end[0] == 'm' && end[1] == 'm') return number * 720 / 254;
+				if (end[0] == 'p' && end[1] == 'c') return number * 12;
+			}
+		}
+	}
+	return number;
+}
+
+int
+svg_parse_enum_from_style(fz_context *ctx, svg_document *doc, const char *style, const char *att,
+	int ecount, const char *etable[], int value)
+{
+	char buf[100], *end, *p;
+	int i;
+	if (style)
+	{
+		p = strstr(style, att);
+		if (p)
+		{
+			size_t n = strlen(att);
+			if (p[n] == ':')
+			{
+				p += n + 1;
+				while (*p && svg_is_whitespace(*p))
+					++p;
+				fz_strlcpy(buf, p, sizeof buf);
+				end = strchr(buf, ';');
+				if (end)
+					*end = 0;
+				for (i = 0; i < ecount; ++i)
+					if (!strcmp(etable[i], buf))
+						return i;
+			}
+		}
+	}
+	return value;
+}
+
+char *
+svg_parse_string_from_style(fz_context *ctx, svg_document *doc, const char *style, const char *att,
+	char *buf, int buf_size, const char *value)
+{
+	char *end, *p, quote;
+	if (style)
+	{
+		p = strstr(style, att);
+		if (p)
+		{
+			size_t n = strlen(att);
+			if (p[n] == ':')
+			{
+				p += n + 1;
+				while (*p && svg_is_whitespace(*p))
+					++p;
+				quote = *p;
+				if (quote == '\'' || quote == '"')
+				{
+					fz_strlcpy(buf, p+1, buf_size);
+					end = strchr(buf, quote);
+				}
+				else
+				{
+					fz_strlcpy(buf, p, buf_size);
+					end = strchr(buf, ';');
+				}
+				if (end)
+					*end = 0;
+				return buf;
+			}
+		}
+	}
+	fz_strlcpy(buf, value, buf_size);
+	return buf;
+}
