@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "xps-imp.h"
 
@@ -116,7 +138,15 @@ static void
 xps_add_link_target(fz_context *ctx, xps_document *doc, char *name)
 {
 	xps_fixpage *page = doc->last_page;
-	xps_target *target = fz_malloc_struct(ctx, xps_target);
+	xps_target *target;
+
+	if (page == NULL)
+	{
+		fz_warn(ctx, "Dropping link target with no page");
+		return;
+	}
+
+	target = fz_malloc_struct(ctx, xps_target);
 
 	fz_try(ctx)
 	{
@@ -290,7 +320,7 @@ xps_parse_metadata(fz_context *ctx, xps_document *doc, xps_part *part, xps_fixdo
 	doc->base_uri = buf;
 	doc->part_uri = part->name;
 
-	xml = fz_parse_xml(ctx, part->data, 0, 0);
+	xml = fz_parse_xml(ctx, part->data, 0);
 	fz_try(ctx)
 	{
 		xps_parse_metadata_imp(ctx, doc, fz_xml_root(xml), fixdoc);
@@ -376,7 +406,7 @@ xps_load_fixed_page(fz_context *ctx, xps_document *doc, xps_fixpage *page)
 	part = xps_read_part(ctx, doc, page->name);
 	fz_try(ctx)
 	{
-		xml = fz_parse_xml(ctx, part->data, 0, 0);
+		xml = fz_parse_xml(ctx, part->data, 0);
 
 		root = fz_xml_root(xml);
 		if (!root)
@@ -431,7 +461,6 @@ static void
 xps_drop_page_imp(fz_context *ctx, fz_page *page_)
 {
 	xps_page *page = (xps_page*)page_;
-	fz_drop_document(ctx, &page->doc->super);
 	fz_drop_xml(ctx, page->xml);
 }
 
@@ -453,13 +482,12 @@ xps_load_page(fz_context *ctx, fz_document *doc_, int chapter, int number)
 			xml = xps_load_fixed_page(ctx, doc, fix);
 			fz_try(ctx)
 			{
-				page = fz_new_derived_page(ctx, xps_page);
+				page = fz_new_derived_page(ctx, xps_page, doc_);
 				page->super.load_links = xps_load_links;
 				page->super.bound_page = xps_bound_page;
 				page->super.run_page_contents = xps_run_page;
 				page->super.drop_page = xps_drop_page_imp;
 
-				page->doc = (xps_document*) fz_keep_document(ctx, (fz_document*)doc);
 				page->fix = fix;
 				page->xml = xml;
 			}
